import logging

import alembic.command
import alembic.config
from sqlmodel import create_engine

from .const import DATABASE_URL, EVAL_ASSIST_DIR, STORAGE_ENABLED
from .model import AppUser, LogRecord, StoredTestCase  # noqa: F401

logger = logging.getLogger(__name__)

engine = None
if STORAGE_ENABLED:
    engine = create_engine(DATABASE_URL)

    alembic_cfg = alembic.config.Config(
        EVAL_ASSIST_DIR / "alembic.ini", config_args={"sqlalchemy.url": DATABASE_URL}
    )
    try:
        alembic.command.upgrade(alembic_cfg, "head")
    except Exception as e:
        logger.exception(f"Error running migrations: {e}")



import json
import logging
import re
from collections.abc import Callable
from typing import Any, cast

from pydantic import BaseModel, ConfigDict, create_model, field_validator

from .types import Criteria, Instance

logger = logging.getLogger(__name__)


def generate_dynamic_pydantic_model(
    model_name: str,
    field_definitions: list[tuple[str, type, Any, list[Callable[..., Any]]]],
) -> type[BaseModel]:
    validators: dict[str, Callable[..., Any]] = {
        validator.__name__: field_validator(field_definition[0], mode="after")(
            validator
        )
        for field_definition in field_definitions
        for validator in field_definition[3]
    }
    field_defs: dict[str, tuple[type, Any]] = {
        field_definition[0]: (field_definition[1], field_definition[2])
        for field_definition in field_definitions
    }
    return create_model(
        model_name,
        __config__=ConfigDict(extra="forbid"),
        __doc__=None,
        __base__=BaseModel,
        __module__=__name__,
        __validators__=validators,
        __cls_kwargs__=None,
        **field_defs,
    )


def get_to_evaluate_text(instance: Instance, criteria: Criteria) -> str | list[str]:
    # if instance only has one field, it must be the text to evaluate
    if len(instance.fields) == 1:
        instance_field = list(instance.fields.keys())[0]
        if criteria.to_evaluate_field != instance_field:
            logger.warning(
                f"Criteria's to_evaluate_field differs from instance's only field. {criteria.to_evaluate_field} != {instance_field}"
            )
        return instance.fields[instance_field]
    else:
        return instance.fields[criteria.to_evaluate_field]


def get_context_dict(instance: Instance, criteria: Criteria) -> dict[str, str]:
    """
    Return a context dict using the instance's fields and the criteria declared context_fields.
    This is useful for multi criteria evaluations where different criteria require different
    contexts and/or text to evaluate.
    """

    # # if instance only has one field, it must be the text to evaluate
    # if len(instance.fields) == 1:
    #     instance_field = list(instance.fields.keys())[0]
    #     if criteria.to_evaluate_field != instance_field:
    #         logger.warning(f"Criteria's to_evaluate_field differs from instance's only field. {criteria.to_evaluate_field} != {instance_field}")
    #     return {}
    if criteria.context_fields is not None:
        # criteria implicitly expects no context
        if len(criteria.context_fields) == 0:
            return {}
        # criteria expects some context, get it from instance.context if available
        if all(field in instance.fields for field in criteria.context_fields):
            return {
                context_field: cast(str, instance.fields[context_field])
                for context_field in criteria.context_fields
            }
        else:
            raise ValueError(
                f"Instance doesn't have the fields that the criteria expects. Criteria's context fields are: {list(criteria.context_fields)} while instance fields are has {list(instance.fields.keys())}"
            )
    # criteria does not specify whether it expects context or not, default to returning the instance non-text-to-evaluate fields as context fields
    return {
        k: cast(str, v)
        for k, v in instance.fields.items()
        if k != criteria.to_evaluate_field
    }


def is_float(element: Any) -> bool:
    # If you expect None to be passed:
    if element is None:
        return False
    try:
        float(element)
        return True
    except ValueError:
        return False


def build_format_instructions(model: type[BaseModel]) -> str:
    """Generate text format instructions based on JSON schema."""
    return (
        "The output should be formatted as a JSON instance that conforms to the JSON schema below.\n\n"
        "As an example, for the schema "
        '{"properties": {"foo": {"title": "Foo", "description": "a list of strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}\n'
        'the object {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. '
        'The object {"properties": {"foo": ["bar", "baz"]}} is not well-formatted.\n\n'
        "Here is the output schema:\n```json\n"
        f"{model.model_json_schema()}\n```\n"
    )


def sanitize_and_parse_json(raw_json: str) -> str:
    """
    Sanitize a JSON string.

    This function:
    - Removes Markdown fences (```json, json```, etc.)
    - Fixes invalid escape sequences (e.g., \', stray backslashes)
    - Escapes unescaped newlines (\n), carriage returns (\r), and tabs (\t)
    - Converts smart quotes to plain quotes
    - Fixes partial or malformed JSON (e.g., missing closing braces)
    - Returns a JSON string that can be safely parsed

    Args:
        raw_json: The raw JSON string (possibly with unescaped characters).

    Returns:
        A sanitized JSON string that can be parsed with json.loads().

    Raises:
        json.JSONDecodeError if the string cannot be fixed.
        TypeError if raw_json is not a string.
    """
    if not isinstance(raw_json, str):
        raise TypeError("raw_json must be a string")

    # -------------------------------
    # 1. Normalize and strip Markdown fences
    # -------------------------------
    raw_json = raw_json.strip()
    markdown_match = re.search(r"```(?:json)?(.*?)```", raw_json, re.DOTALL)
    json_str = markdown_match.group(1).strip() if markdown_match else raw_json

    # -------------------------------
    # 2. Normalize invalid characters and escapes
    # -------------------------------
    # Replace smart quotes and backticks
    json_str = (
        json_str.replace("“", '"')
        .replace("”", '"')
        .replace("‘", "'")
        .replace("’", "'")
        .replace("`", '"')
    )

    # Replace Python-style escaped single quotes with plain quotes
    json_str = re.sub(r"\\'", "'", json_str)

    # Remove invalid backslash escapes (anything not valid JSON escape)
    json_str = re.sub(r'\\(?!["\\/bfnrtu])', r"\\\\", json_str)

    # -------------------------------
    # 3. Escape problematic characters inside string values
    # -------------------------------
    def _replace_chars(match: re.Match[str]) -> str:
        value = match.group(2)
        value = re.sub(r"\n", r"\\n", value)
        value = re.sub(r"\r", r"\\r", value)
        value = re.sub(r"\t", r"\\t", value)
        value = re.sub(r'(?<!\\)"', r"\"", value)
        return match.group(1) + value + match.group(3)

    json_str = re.sub(
        r'(".*?"\s*:\s*")(.*?)(")', _replace_chars, json_str, flags=re.DOTALL
    )

    # -------------------------------
    # 4. Try parsing and, if needed, repair partial JSON
    # -------------------------------
    try:
        json.loads(json_str)
        return json_str
    except json.JSONDecodeError:
        if not json_str:
            raise json.JSONDecodeError("Empty string is not valid JSON", json_str, 0)
        new_chars = []
        stack = []
        is_inside_string = False
        escaped = False

        for char in json_str:
            new_char = char
            if is_inside_string:
                if char == '"' and not escaped:
                    is_inside_string = False
                elif char == "\n" and not escaped:
                    new_char = "\\n"
                elif char == "\\":
                    escaped = not escaped
                else:
                    escaped = False
            elif char == '"':
                is_inside_string = True
                escaped = False
            elif char == "{":
                stack.append("}")
            elif char == "[":
                stack.append("]")
            elif char in {"}", "]"}:
                if stack and stack[-1] == char:
                    stack.pop()
                else:
                    return ""
            new_chars.append(new_char)

        if is_inside_string:
            if escaped:
                new_chars.pop()
            new_chars.append('"')

        stack.reverse()

        while new_chars:
            candidate = "".join(new_chars + stack)
            try:
                json.loads(candidate)
                return candidate
            except json.JSONDecodeError:
                new_chars.pop()

        return json_str


def create_default_instance(
    model_cls: type[BaseModel], values: dict[str, Any] | None = None
) -> BaseModel:
    """
    Create an instance of a Pydantic model with default placeholder values.
    If 'values' is provided, it overrides the defaults for matching fields.
    """
    values = values or {}
    defaults: dict[str, Any] = {}

    for name, field in model_cls.model_fields.items():
        if name in values:
            defaults[name] = values[name]
        else:
            defaults[name] = ""

    return model_cls(**defaults)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, cast

from evalassist.judges import Criteria, CriteriaOption
from fastapi import HTTPException
from pydantic import BaseModel, Field, RootModel, field_validator
from unitxt.llm_as_judge import EvaluatorNameEnum, EvaluatorTypeEnum, ModelProviderEnum

from .extended_unitxt import ExtendedEvaluatorNameEnum, ExtendedModelProviderEnum
from .judges.types import (
    DirectInstanceResult,
    Instance,
    InstanceResult,
    PairwiseInstanceResult,
)
from .model import StoredTestCase


class TaskEnum(str, Enum):
    SUMMARIZATION = "Summarization"
    TEXT_GENERATION = "Text Generation"
    QUESTION_ANSWERING = "Question Answering"


class DomainEnum(str, Enum):
    NEWS_MEDIA_DOMAIN = "News Media"
    HEALTHCARE = "Healthcare"
    ENTERTAINMENT_AND_POP_CULTURE = "Entertainment And Pop Culture"

    SOCIAL_MEDIA = "Social Media"
    CUSTOMER_SUPPORT_AND_BUSSINESS = "Custumer Support And Business"
    GAMING_AND_ENTERTAINMENT = "Gaming And Entertainment"


class PersonaEnum(str, Enum):
    EXPERIENCED_JOURNALIST = "Experienced journalist"
    NOVICE_JOURNALIST = "Novice journalist"
    OPINION_COLUMNIST = "Opinion columnist"
    NEWS_ANCHOR = "News anchor"
    EDITOR = "Editor"

    MEDICAL_RESEARCHER = "Medical researcher"
    GENERAL_PRACTITIONER = "General practitioner"
    PUBLIC_HEALTH_OFFICIAL = "Public health official"
    HEALTH_BLOGGER = "Health blogger"
    MEDICAL_STUDENT = "Medical student"

    FILM_CRITIC = "Film critic"
    CASUAL_SOCIAL_MEDIA_USER = "Casual social media user"
    TABLOID_REPORTER = "Tabloid reporter"
    HARDCORE_FAN_THEORIST = "Hardcore fan/Theorist"
    INFLUENCER_YOUTUBE_REVIEWER = "Inlfuencer/Youtube reviewer"

    INFLUENCER_POSITIVE_BRAND = "Influencer (Positive brand)"
    INTERNET_TROLL = "Internet troll"
    POLITICAL_ACTIVIST = "Political activist (polarizing)"
    BRAND_VOICE = "Brand voice (Corporate social media account)"
    MEMER = "Memer (Meme creator)"
    CUSTOMER_SERVICE_AGENT = "Customer service agent"
    ANGRY_CUSTOMER = "Angry customer"
    CORPORATE_CEO = "Corporate CEO"
    CONSUMER_ADVOCATE = "Consumer advocate"
    MAKETING_SPECIALIST = "Marketing specialist"

    FLAMER = "Flamer (Agressive player)"
    HARDCORE_GAMER = "Hardcore gamer"
    ESPORT_COMENTATOR = "Esport commentator"
    MOVIE_CRITIC = "Movie critic"
    FAN = "Fan (of a TV show, movie, or game)"


class GenerationLengthEnum(str, Enum):
    SHORT = "Short"
    MEDIUM = "Medium"
    LONG = "Long"


class DirectActionTypeEnum(str, Enum):
    REGENERATE = "Regenerate"
    REPHRASE = "Rephrase"
    LONGER = "Elaborate"
    SHORTER = "Shorten"
    CUSTOM = "Custom"


class CriteriaDTO(BaseModel):
    name: str
    description: str
    to_evaluate_field: str
    context_fields: list[str]

    @field_validator("description")
    def validate_criteria(cls, description):
        if len(description.strip()) == 0:
            raise HTTPException(
                status_code=400, detail="Evaluation criteria description is required."
            )
        return description

    def to_criteria(self, examples: list[InstanceResult]) -> Criteria:
        return Criteria(
            name=self.name,
            description=self.description,
            to_evaluate_field=self.to_evaluate_field,
            context_fields=self.context_fields,
            examples=examples,
        )


class CriteriaOptionDTO(BaseModel):
    name: str
    description: str


class CriteriaWithOptionsDTO(CriteriaDTO):
    name: str
    description: str
    options: list[CriteriaOptionDTO]

    def to_criteria(self, examples: list[InstanceResult]) -> Criteria:
        return Criteria(
            name=self.name,
            description=self.description,
            options=[
                CriteriaOption(name=o.name, description=o.description)
                for o in self.options
            ],
            to_evaluate_field=self.to_evaluate_field,
            context_fields=self.context_fields,
            examples=examples,
        )


class InstanceDTO(BaseModel, ABC):
    id: str
    metadata: dict[str, Any] | None = Field(
        default=None,
        description="Additional metadata about the instance, stored as key-value pairs.",
    )
    context: dict
    expected_result: str | int | None = Field(
        default=None,
        description="The expected result or ground truth for the instance, if available.",
    )

    def to_instance(self, to_evaluate_field: str) -> Instance:
        return Instance(
            fields={
                **(self.context if self.context is not None else {}),
                to_evaluate_field: self.get_prediction(),
            }
        )

    def to_instance_result(self, to_evaluate_field: str) -> InstanceResult:
        return InstanceResult(
            instance=self.to_instance(to_evaluate_field),
            selected_option=cast(str, self.expected_result),
        )

    @abstractmethod
    def get_prediction(self) -> Any: ...  # noqa: E704


class DirectInstanceDTO(InstanceDTO):
    response: str = Field(
        description="The response or prediction generated for the instance."
    )

    def get_prediction(self):
        return self.response


class PairwiseInstanceDTO(InstanceDTO):
    responses: list[str] = Field(
        description="A list of responses or predictions to be compared in a pairwise evaluation."
    )

    def get_prediction(self):
        return self.responses


class EvaluationRequest(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: str
    type: EvaluatorTypeEnum
    instances: list[DirectInstanceDTO] | list[PairwiseInstanceDTO]
    criteria: CriteriaDTO | CriteriaWithOptionsDTO
    examples: list[DirectInstanceDTO] | list[PairwiseInstanceDTO]
    judge: str
    judge_params: dict


class PairwiseInstanceResultDTO(BaseModel):
    id: str
    result: PairwiseInstanceResult


class PairwiseResultDTO(BaseModel):
    results: list[PairwiseInstanceResultDTO]


class NotebookParams(BaseModel):
    test_case_name: str
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum
    provider: ModelProviderEnum
    credentials: dict[str, str]
    evaluator_type: EvaluatorTypeEnum
    plain_python_script: bool
    instances: list[DirectInstanceDTO] | list[PairwiseInstanceDTO]
    examples: list[DirectInstanceDTO] | list[PairwiseInstanceDTO]
    criteria: CriteriaDTO | CriteriaWithOptionsDTO
    judge: str
    judge_params: dict


# class DownloadTestCaseParams(BaseModel):
#     test_case: TestCase


class SyntheticExampleGenerationRequest(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str
    type: EvaluatorTypeEnum
    criteria: CriteriaWithOptionsDTO | CriteriaDTO
    generation_length: GenerationLengthEnum | None
    task: TaskEnum | None
    domain: DomainEnum | None
    persona: PersonaEnum | None
    per_criteria_option_count: dict[str, int]
    borderline_count: int


class DirectInstanceResultDTO(BaseModel):
    id: str
    result: DirectInstanceResult


class InstanceResultWithId(BaseModel):
    id: str
    result: DirectInstanceResult | PairwiseInstanceResult


class DirectResultDTO(BaseModel):
    results: list[DirectInstanceResultDTO]


class InstanceResultDTO(RootModel):
    root: DirectInstanceResultDTO | PairwiseInstanceResultDTO


class EvaluationResultDTO(BaseModel):
    results: list[InstanceResultWithId]


class TestModelRequestModel(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str


class DirectAIActionRequest(BaseModel):
    action: DirectActionTypeEnum
    selection: str
    text: str
    prompt: str | None = None
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str
    type: EvaluatorTypeEnum


class DirectAIActionResponse(BaseModel):
    result: str


class FeatureFlagsModel(BaseModel):
    authentication_enabled: bool
    storage_enabled: bool


class PutTestCaseBody(BaseModel):
    user: str
    test_case: StoredTestCase


class DownloadTestCaseBody(BaseModel):
    test_case: StoredTestCase


class DownloadTestDataBody(BaseModel):
    instances: list[DirectInstanceDTO] | list[PairwiseInstanceDTO]
    to_evaluate_field: str


class EvaluatorMetadataAPI(BaseModel):
    name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str
    providers: list[ModelProviderEnum | ExtendedModelProviderEnum]


class EvaluatorsResponseModel(BaseModel):
    # model_config = ConfigDict(arbitrary_types_allowed=True)
    evaluators: list[EvaluatorMetadataAPI]


class JudgesResponseModel(BaseModel):
    judges: dict[str, list]


class FixInstanceRequest(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: str
    type: EvaluatorTypeEnum
    instance: DirectInstanceDTO
    result: DirectInstanceResult


class FixInstanceResponse(BaseModel):
    fixed_response: str



from .base import BaseDirectJudge, BaseJudge, BasePairwiseJudge
from .const import DEFAULT_JUDGE_INFERENCE_PARAMS
from .direct_judge import DirectJudge
from .dummy_judge import DummyDirectJudge, DummyPairwiseJudge
from .experimental.criticized_judge import CriticizedDirectJudge
from .experimental.thesis_antithesis_direct_judge import ThesisAntithesisDirectJudge
from .mprometheus_judge import MPrometheusDirectJudge, MPrometheusPairwiseJudge
from .pairwise_judge import PairwiseJudge
from .types import (
    Criteria,
    CriteriaOption,
    DirectInstanceResult,
    DirectPositionalBiasResult,
    Instance,
    MultiCriteria,
    MultiCriteriaDirectInstanceResult,
    MultiCriteriaItem,
    PairwiseInstanceResult,
    PairwisePositionalBiasResult,
    SingleSystemPairwiseInstanceResult,
)
from .unitxt_judges import GraniteGuardianJudge, UnitxtDirectJudge, UnitxtPairwiseJudge

JUDGE_CLASS_MAP = {
    "direct": {
        "eval_assist": DirectJudge,
        "m_prometheus": MPrometheusDirectJudge,
        "criticized": CriticizedDirectJudge,
        "thesis_antithesis": ThesisAntithesisDirectJudge,
        "unitxt": UnitxtDirectJudge,
        "granite_guardian": GraniteGuardianJudge,
    },
    "pairwise": {
        "eval_assist": PairwiseJudge,
        "m_prometheus": MPrometheusPairwiseJudge,
        "unitxt": UnitxtPairwiseJudge,
    },
}

# BASE_PARAMS = {
#     "check_positional_bias": False,
#     "self_consistency": int
# }

# JUDGE_PARAMS_MAP = {
#     'direct': {
#         'eval_assist': {
#             'generate_synthetic_persona': bool,
#             'generate_feedback': bool,
#             'on_generation_failure': ["raise", "random"]
#         },
#         'm_prometheus': {},
#         'criticized': {},
#         'thesis_antithesis': {},
#         'unitxt': {},
#         'granite_guardian': {},
#     },
#     'pairwise': {
#         'eval_assist': {},
#         'm_prometheus': {},
#         'unitxt': {},
#     }
# }


JUDGES_CONFIG = {}

__all__: list[str] = [
    "BaseJudge",
    "DummyDirectJudge",
    "DummyPairwiseJudge",
    "DirectJudge",
    "PairwiseJudge",
    "UnitxtDirectJudge",
    "UnitxtPairwiseJudge",
    "BaseDirectJudge",
    "BasePairwiseJudge",
    "Instance",
    "SingleSystemPairwiseInstanceResult",
    "PairwiseInstanceResult",
    "DirectPositionalBiasResult",
    "PairwisePositionalBiasResult",
    "DirectInstanceResult",
    "DEFAULT_JUDGE_INFERENCE_PARAMS",
    "Criteria",
    "CriteriaOption",
    "MPrometheusDirectJudge",
    "MPrometheusPairwiseJudge",
    "MultiCriteria",
    "MultiCriteriaItem",
    "MultiCriteriaDirectInstanceResult",
    "GraniteGuardianJudge",
]



"""Init

Revision ID: 6da8778318cb
Revises:
Create Date: 2025-06-27 14:03:12.997706

"""

from typing import Sequence

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "6da8778318cb"  # pragma: allowlist secret
down_revision: str | Sequence[str] | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "app_user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("email", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_app_user_email"), "app_user", ["email"], unique=True)
    op.create_table(
        "log_record",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("data", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "stored_test_case",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("content", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["app_user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("stored_test_case")
    op.drop_table("log_record")
    op.drop_index(op.f("ix_app_user_email"), table_name="app_user")
    op.drop_table("app_user")
    # ### end Alembic commands ###



"""Add user_id to log_record

Revision ID: 004150514536
Revises: 6da8778318cb
Create Date: 2025-06-27 14:27:41.163987

"""

from typing import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "004150514536"  # pragma: allowlist secret
down_revision: str | Sequence[str] | None = "6da8778318cb"  # pragma: allowlist secret
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("log_record", schema=None) as batch_op:
        batch_op.add_column(sa.Column("user_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            "fk_log_record_user_id", "app_user", ["user_id"], ["id"]
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("log_record", schema=None) as batch_op:
        batch_op.drop_constraint("fk_log_record_user_id", type_="foreignkey")
        batch_op.drop_column("user_id")

    # ### end Alembic commands ###



import logging

import click
import uvicorn
from evalassist.const import UVICORN_WORKERS
from evalassist.utils import get_system_version

logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@cli.command()
@click.option("--host", default="127.0.0.1", help="Host to bind to.")
@click.option("--port", default=8000, type=int, help="Port to bind to.")
@click.option("--reload", default=False, type=bool, help="Reload on changes.")
def serve(host: str, port: int, reload: bool):
    logger.info(
        f"Starting EvalAssist on {'http://' if not host.startswith('http') else ''}{host}:{port}..."
    )
    uvicorn.run(
        "evalassist.main:app",
        host=host,
        port=port,
        loop="asyncio",
        reload=reload,
        workers=UVICORN_WORKERS,
        log_level="debug",
    )


@cli.command()
def version():
    version_info = get_system_version()
    click.echo(f"EvalAssist {version_info['version']} (from {version_info['source']})")


def main():
    cli()


if __name__ == "__main__":
    cli()



import json
import logging
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import cast

from datasets import IterableDataset
from evalassist.judges import Criteria
from evalassist.judges.types import InstanceResult
from unitxt.api import evaluate, load_dataset
from unitxt.artifact import fetch_artifact
from unitxt.llm_as_judge import CriteriaWithOptions

from ..const import EVAL_ASSIST_DIR
from ..judges import BaseDirectJudge
from ..judges.types import DirectInstanceResult, Instance
from ..utils import convert_nan_to_none, unitxt_dataset_to_evalassist_instances
from .utils import (
    add_judgebench_readme_urls,
    add_tag_to_result,
    add_url_to_result,
    get_benchmark_results_as_df,
    get_judge_from_config,
    get_judgebench_cards,
    save_evaluation_backup_to_sqlite,
    save_results_to_sqlite,
)

RESULTS_FILE_PATH = EVAL_ASSIST_DIR / "benchmark" / "benchmark_results.csv"
CACHE_FILE_PATH = EVAL_ASSIST_DIR / "benchmark" / "benchmark_results_cache.csv"

logger = logging.getLogger(__name__)


def get_all_benchmarks():
    df = get_benchmark_results_as_df()
    benchmarks = {}
    for row in df.to_dict(orient="records"):
        card: str = row["card"]
        benchmark_name: str = row["benchmark_name"]
        dataset_name: str = row["dataset_name"]
        judge: str = row["judge"]
        model: str = row["model"]
        results: str = row["results"]
        dataset_len: str = row["dataset_len"]
        group_by_field: str = (
            row["group_by_field"] if row["group_by_value"] is not None else "overall"
        )
        group_by_value: str = (
            row["group_by_value"] if row["group_by_field"] is not None else "overall"
        )
        benchmark_id = benchmark_name + "/" + dataset_name

        if benchmark_id not in benchmarks:
            benchmark_results = {
                "benchmark_name": benchmark_name,
                "dataset_name": dataset_name,
                "display_name": (
                    benchmark_name + "." if benchmark_name != "judge_bench" else ""
                )
                + dataset_name,  # don't include the benchmark name in the display name if it is judge_bench
                "description": "",
                "catalog_url": f"https://www.unitxt.ai/en/latest/catalog/catalog.{card}.html",
                "type": "direct",
                "tags": [benchmark_name],
                "group_by_fields": {},
            }

            benchmarks[benchmark_id] = benchmark_results

        benchmark_results = benchmarks[benchmark_id]

        if group_by_field not in benchmark_results["group_by_fields"]:
            benchmark_results["group_by_fields"][group_by_field] = {}

        group_by_field_results = benchmark_results["group_by_fields"][group_by_field]

        if group_by_value not in group_by_field_results:
            group_by_value_results = {
                "judge_results": {},
                "dataset_len": dataset_len,
                "group_by_value": group_by_value,
            }
            group_by_field_results[group_by_value] = group_by_value_results

        group_by_value_results = group_by_field_results[group_by_value]
        judge_model_key = f"{judge}/{model}"
        group_by_value_results["judge_results"][judge_model_key] = {
            "model": model,
            "judge": judge,
            "results": json.loads(results),
        }

    # add benchmark to the name if it only has one dataset
    datasets_per_benchmarks = {}
    for r in benchmarks.values():
        if r["benchmark_name"] not in datasets_per_benchmarks:
            datasets_per_benchmarks[r["benchmark_name"]] = []
        if r["dataset_name"] not in datasets_per_benchmarks[r["benchmark_name"]]:
            datasets_per_benchmarks[r["benchmark_name"]].append(r["dataset_name"])

    add_judgebench_readme_urls(benchmarks)
    add_tag_to_result(benchmarks, "roscoe", "reasoning")
    add_tag_to_result(benchmarks, "wmt", "translation")
    add_tag_to_result(benchmarks, "cola", "grammar")

    add_url_to_result(
        benchmarks,
        "biggen",
        "https://huggingface.co/datasets/prometheus-eval/BiGGen-Bench-Results/viewer/default/human_eval",
    )

    return benchmarks


metric_map = {
    "pearson": "pearsonr",
    "spearman": "spearmanr",
}


def parse_and_store_results(
    card: str,
    dataset: IterableDataset,
    instances: list[Instance],
    results: list[DirectInstanceResult],
    judge: BaseDirectJudge,
    add_in_context_examples: bool,
):
    judge_descriptor = judge.get_descriptor()
    judge_name = judge_descriptor.name + (
        "_with_examples" if add_in_context_examples else ""
    )
    criteria_names = []
    for result in results:
        if result.criteria is not None and result.criteria.name is None:
            raise ValueError(
                "Criteria doesn't have a name. When benchmarking, it is required."
            )
        criteria_names.append(cast(str, cast(Criteria, result.criteria).name))

    if any(result.score is None for result in results):
        raise ValueError("Score is None!")

    prediction_scores: list[float] = [cast(float, result.score) for result in results]

    str_prediction_scores = [str(p) for p in prediction_scores]

    # Calculate positional bias rate
    positional_bias_rate = sum(
        [
            r.positional_bias.detected if r.positional_bias is not None else False
            for r in results
        ]
    ) / len(results)

    evaluation_results = evaluate(predictions=str_prediction_scores, data=dataset)

    # Extract metric names from the evaluation results
    metric_names = [m.split(".")[1] for m in evaluation_results[0]["metrics"]]

    # Parse the evaluation results into a dictionary
    parsed_metric_results = {
        metric_name: float(
            evaluation_results.global_scores[metric_map.get(metric_name, metric_name)]
        )
        for metric_name in metric_names
    }

    # Store the positional bias rate in the parsed results
    parsed_metric_results["positional_bias_rate"] = positional_bias_rate

    benchmark_name = card.split(".")[1]
    dataset_name = ".".join(
        card.split(".")[2:-1]
        if benchmark_name.startswith("judge_bench")
        else card.split(".")[2:]
    )
    benchmark_criteria_name = (
        card.split(".")[-1]
        if benchmark_name.startswith("judge_bench")
        else criteria_names[0]
    )
    benchmark_result: dict[str, str | None] = {
        "card": card,  # if there are several criteria, we have to add the overall result
        "benchmark_name": benchmark_name,
        "dataset_name": dataset_name,
        "judge": judge_name,
        "model": judge_descriptor.inference_engine_id,
        "results": json.dumps(convert_nan_to_none(parsed_metric_results)),
        "dataset_len": str(evaluation_results.global_scores["num_of_instances"]),
        "group_by_field": None,
        "group_by_value": None,
    }

    benchmark_results: list = []
    benchmark_results.append(benchmark_result)
    for (
        group_by_field,
        group_by_field_scores,
    ) in evaluation_results.groups_scores.items():
        for group_by_value, group_by_value_scores in group_by_field_scores.items():
            parsed_metric_results: dict[str, float | None] = {
                metric_name: float(
                    group_by_value_scores[metric_map.get(metric_name, metric_name)]
                )
                for metric_name in metric_names
            }
            parsed_metric_results["positional_bias_rate"] = None
            group_by_field_corrected = (
                group_by_field if group_by_field != "criteria/name" else "criteria"
            )
            per_criteria_benchmark_result: dict[str, str | None] = {
                "card": card,  # if there are several criteria, we have to add the overall result
                "benchmark_name": benchmark_name,
                "dataset_name": dataset_name,
                "judge": judge_name,
                "model": judge_descriptor.inference_engine_id,
                "results": json.dumps(convert_nan_to_none(parsed_metric_results)),
                "dataset_len": str(group_by_value_scores["num_of_instances"]),
                "group_by_field": group_by_field_corrected,
                "group_by_value": group_by_value,
            }
            if (
                group_by_field_corrected == "criteria"
                and benchmark_name == " biggen_bench"
            ):
                per_criteria_benchmark_result["group_by_value"] = cast(
                    str, per_criteria_benchmark_result["group_by_value"]
                ).split("-")[0]  # we leave the task and discard the capability
            benchmark_results.append(per_criteria_benchmark_result)

    ground_truth = [float(d["target"]) for d in dataset]
    result_backup: list[dict] = [
        {
            "card": card,
            "judge": judge_name,
            "model": judge_descriptor.inference_engine_id,
            "sample_index": i,
            "instance": instance.model_dump_json(indent=4),
            "result": result.model_dump_json(indent=4),
            "prediction": prediction,
            "ground_truth": target,
            "failed": prediction != target,
        }
        for i, (prediction, target, instance, result) in enumerate(
            zip(prediction_scores, ground_truth, instances, results)
        )
    ]

    # there is only one grouped by result: so the list contains the overall result and a single criteria result, we keep the later
    # and set the benchmark criteria name from the card name
    if len(benchmark_results) == 2:
        benchmark_results = [benchmark_results[1]]
        benchmark_criteria_name = (
            card.split(".")[-1]
            if benchmark_name.startswith("judge_bench")
            else criteria_names[0]
        )
        benchmark_results[0]["group_by_value"] = benchmark_criteria_name

    save_evaluation_backup_to_sqlite(result_backup)
    save_results_to_sqlite(benchmark_results)


def add_examples(
    instances: list[Instance], criteria: list[Criteria], ground_truth: list[float]
):
    d = {}
    for instance, criterion, label in zip(instances, criteria, ground_truth):
        if criterion.name not in d:
            d[criterion.name] = {
                "criteria": criterion,
                "instances": [],
                "ground_truth": [],
            }
        if label not in d[criterion.name]["ground_truth"]:
            d[criterion.name]["instances"].append(instance)
            d[criterion.name]["ground_truth"].append(label)
    for k, v in d.items():
        v["examples"] = [
            InstanceResult(
                instance=instance,
                selected_option=next(
                    iter(
                        option.name
                        for option in v["criteria"].options
                        if option.score == ground_truth
                    )
                ),
            )
            for instance, ground_truth in zip(v["instances"], v["ground_truth"])
        ]
    for criterion in criteria:
        criterion.examples = d[criterion.name]["examples"]


def run_single_model_card(
    card: str,
    judge: BaseDirectJudge,
    add_in_context_examples: bool,
    instances_per_dataset: int | None = None,
):
    """
    Runs a single benchmark card with the specified model and API key.

    Args:
        card (str): The name of the benchmark card to run.
        dataset: The dataset to use for benchmarking.
        model (str): The name of the model to use for benchmarking.
        api_key (str): The API key to use for the model.

    Returns:
        tuple: A tuple containing the benchmark result and inspection rows.
    """
    print(
        "Running card:",
        card,
        "with judge:",
        judge.get_descriptor(),
    )

    try:
        group_by_fields = ["criteria/name"]
        if "biggen" in card:
            group_by_fields = group_by_fields + ["language", "capability"]
        try:
            dataset: IterableDataset = cast(
                IterableDataset,
                load_dataset(
                    card=card,
                    split="test",
                    loader_limit=instances_per_dataset,
                    use_cache=True,
                    group_by=group_by_fields,
                ),
            )
        except ValueError:
            group_by_fields[0] = "criteria"
            dataset: IterableDataset = cast(
                IterableDataset,
                load_dataset(
                    card=card,
                    split="test",
                    loader_limit=instances_per_dataset,
                    use_cache=True,
                    group_by=group_by_fields,
                ),
            )
        criteria: list[Criteria] = [
            Criteria.from_unitxt_criteria(
                cast(
                    CriteriaWithOptions,
                    fetch_artifact(json.loads(d["task_data"])["criteria"])[0],
                )
            )
            for d in dataset
        ]

        parsed_dataset: list[Instance] = unitxt_dataset_to_evalassist_instances(
            dataset, criteria
        )

        ground_truth = [float(d["target"]) for d in dataset]

        if add_in_context_examples:
            add_examples(parsed_dataset, criteria, ground_truth)

        results: list[DirectInstanceResult] = judge(
            parsed_dataset,
            criteria,
        )

        parse_and_store_results(
            card=card,
            dataset=dataset,
            instances=parsed_dataset,
            results=results,
            judge=judge,
            add_in_context_examples=add_in_context_examples,
        )
        print("Finished running card:", card, "with judge:", judge.get_descriptor())
    except Exception:
        print(f"FAILED! judege: {str(judge.get_descriptor())}")
        print(traceback.format_exc())


def run_benchmarks(
    judge_configs: list[tuple[type[BaseDirectJudge], dict, dict, str]],
    max_workers: int,
    instances_per_dataset: int | None,
    add_in_context_examples: bool = False,
    dataset_keyword_filters: list[str] | None = None,
    dataset_keyword_selectors: list[str] | None = None,
):
    """

    Runs multiple benchmarks in parallel using a process pool executor.

    This function retrieves a list of JudgeBench cards, loads the corresponding datasets,
    and then submits tasks to the executor to run each benchmark with different models.

    The results are saved to CSV files specified by RESULTS_FILE_PATH and INSPECT_FILE_PATH.
    """

    # Create a cycle of API keys to use for benchmarking
    all_benchmarks = [
        "cards.biggen_bench.results.human_eval",
    ] + get_judgebench_cards()

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for card in all_benchmarks:
            # Load the dataset for the current card
            if dataset_keyword_filters is not None and any(
                x in card for x in dataset_keyword_filters
            ):
                continue
            if dataset_keyword_selectors is not None and not any(
                x in card for x in dataset_keyword_selectors
            ):
                continue
            for judge_config in judge_configs:
                judge = get_judge_from_config(judge_config)

                # Skip if the benchmark has already been run
                # if not is_result_available(
                #     judge.get_name(),
                #     judge.get_descriptor().inference_engine_id,
                #     card,
                # ):
                if True:
                    # if True:
                    # Submit the task to the executor
                    futures.append(
                        executor.submit(
                            run_single_model_card,
                            card,
                            judge,
                            add_in_context_examples,
                            instances_per_dataset,
                        )
                    )
                else:
                    print(
                        f"Benchmark {card}/{judge.get_descriptor()}/{judge.get_name()} already ran"
                    )
    print("Done running benchmarks")



import datetime

from sqlalchemy import Column, DateTime, func
from sqlmodel import Field, Relationship, SQLModel


class AppUser(SQLModel, table=True):
    __tablename__ = "app_user"  # type: ignore
    id: int | None = Field(default=None, primary_key=True)
    email: str = Field(index=True, unique=True)
    name: str
    created_at: datetime.datetime | None = Field(
        default=None,
        sa_column=Column(DateTime(), server_default=func.now(), nullable=False),
    )
    stored_test_cases: list["StoredTestCase"] = Relationship(back_populates="app_user")
    log_records: list["LogRecord"] = Relationship(back_populates="app_user")


class StoredTestCase(SQLModel, table=True):
    __tablename__ = "stored_test_case"  # type: ignore
    id: int | None = Field(default=None, primary_key=True)
    user_id: int = Field(foreign_key="app_user.id")
    content: str
    name: str

    app_user: AppUser | None = Relationship(back_populates="stored_test_cases")


class LogRecord(SQLModel, table=True):
    __tablename__ = "log_record"  # type: ignore
    id: int | None = Field(default=None, primary_key=True)
    data: str
    user_id: int = Field(foreign_key="app_user.id")

    app_user: AppUser | None = Relationship(back_populates="log_records")



import json
import logging
import random
from typing import Literal, cast

from evalassist.judges.utils import create_default_instance, sanitize_and_parse_json
from pydantic import BaseModel, ValidationError
from unitxt.inference import InferenceEngine

logger = logging.getLogger(__name__)


class BatchRepairParser:
    def __init__(
        self,
        inference_engine: InferenceEngine,
        max_retries: int = 3,
        on_generation_failure: Literal["raise", "random"] = "random",
    ):
        """
        inference_engine: object with `infer(dataset: list[dict]) -> list[str]`
        max_retries: number of repair rounds
        on_generation_failure: "raise" or "random"
        """
        self.inference_engine: InferenceEngine = inference_engine
        self.max_retries = max_retries
        self.on_generation_failure = on_generation_failure

    def parse_and_repair(
        self,
        unparsed_responses: list[str],
        model_classes: list[type[BaseModel]],
        on_failure_default: list[dict] | dict | None = None,
    ) -> tuple[list[BaseModel], list[dict]]:
        """
        Parse all responses into Pydantic models, repairing failed responses in batches.
        """
        if isinstance(on_failure_default, dict):
            on_failure_default = [on_failure_default] * len(unparsed_responses)
        n = len(unparsed_responses)
        sanitized_unparsed_responses = [
            sanitize_and_parse_json(x) for x in unparsed_responses
        ]
        parsed: list[BaseModel | None] = [None] * n
        metadata: list[dict] = [{} for _ in range(n)]

        # Initial parse attempt
        failures = []
        for i, (text, model_class) in enumerate(
            zip(sanitized_unparsed_responses, model_classes)
        ):
            try:
                parsed[i] = model_class.model_validate_json(text, strict=False)
                metadata[i] = {"generation_failed": False}
            except ValidationError as e:
                failures.append(i)
                metadata[i] = {
                    "generation_failed": True,
                    "generation_failed_original_output": text,
                    "parsing_error": str(e),
                }

        attempt = 0
        while failures and attempt < self.max_retries:
            attempt += 1
            logger.debug(
                f"BatchRepairParser: repair attempt {attempt}/{self.max_retries} for {len(failures)} items (out of {len(sanitized_unparsed_responses)} items) using inference engine {self.inference_engine.get_engine_id()}"
            )
            logger.debug(
                f"First incorrect unparsed response is:\n{sanitized_unparsed_responses[failures[0]]}"
            )
            logger.debug(
                f"First parsing issue was:\n{metadata[failures[0]].get('parsing_error', '')}"
            )

            dataset = []
            idx_map = []

            for idx in failures:
                raw_text = sanitized_unparsed_responses[idx]
                model_class = model_classes[idx]
                prompt_text = self._format_repair_prompt(
                    invalid_output=raw_text,
                    parsing_error=metadata[idx].get("parsing_error", ""),
                    model_class=model_class,
                )

                dataset.append(
                    {
                        "source": prompt_text,
                        "data_classification_policy": ["public"],
                    }
                )
                idx_map.append(idx)

            # Send batch to inference engine
            try:
                repaired_texts = [
                    sanitize_and_parse_json(str(r))
                    for r in self.inference_engine.infer(dataset=dataset)
                ]
            except Exception as e:
                logger.exception(
                    "BatchRepairParser: inference_engine.infer failed on retry %s", e
                )
                break

            # Attempt parsing repaired responses
            new_failures = []
            for pos_in_batch, repaired_text in enumerate(repaired_texts):
                original_index = idx_map[pos_in_batch]
                model_class = model_classes[original_index]

                sanitized_unparsed_responses[original_index] = repaired_text
                metadata[original_index]["generation_failed_last_attempt_output"] = (
                    repaired_text
                )
                metadata[original_index]["repair_attempts"] = attempt

                try:
                    parsed_obj = model_class.model_validate_json(repaired_text)
                    parsed[original_index] = parsed_obj
                    metadata[original_index]["generation_failed"] = False
                except ValidationError as e:
                    metadata[original_index]["parsing_error"] = str(e)
                    new_failures.append(original_index)

            failures = new_failures

        if failures:
            logger.debug(
                f"BatchRepairParser: repairing failed for {len(failures)} items. Applying on_generation_failure='{self.on_generation_failure}'."
            )

        # Fallback for remaining failures
        for i in failures:
            if self.on_generation_failure == "raise":
                raise ValueError(
                    f"Failed to parse response after {attempt} attempts.\n"
                    f"Original output:\n{metadata[i].get('generation_failed_original_output')}\n"
                    f"Last attempt:\n{metadata[i].get('generation_failed_final_output')}\n"
                    f"Error:\n{metadata[i].get('parsing_error')}"
                )

            if on_failure_default is not None:
                defaults: dict = on_failure_default[i]
                for k, v in defaults.items():
                    if isinstance(v, list):
                        defaults[k] = random.choice(v)  # nosec

                model_class = model_classes[i]
                parsed[i] = create_default_instance(model_class, defaults)

                metadata[i]["generation_failed"] = True
                metadata[i]["final_fallback_chosen"] = defaults

        return cast(list[BaseModel], parsed), metadata

    def _format_repair_prompt(
        self,
        invalid_output: str,
        parsing_error: str,
        model_class: type[BaseModel],
    ) -> str:
        """
        Build a repair prompt using the Pydantic model's JSON schema.
        """
        try:
            schema_json = json.dumps(
                model_class.model_json_schema(), indent=2, ensure_ascii=False
            )
        except Exception:
            schema_json = "<unable to generate model schema>"

        prompt = (
            "You are given an invalid model output that must be corrected "
            "to match the expected JSON schema.\n\n"
            f"INVALID OUTPUT:\n```\n{invalid_output}\n```\n\n"
            f"PARSING ERROR:\n```\n{parsing_error}\n```\n\n"
            f"EXPECTED JSON SCHEMA:\n{schema_json}\n\n"
            "INSTRUCTIONS:\n"
            "1) Produce a single JSON object strictly conforming to the schema above.\n"
            "2) Do NOT include explanations or extra text.\n"
            "3) If a field's value is unknown, use empty string, null, or logical empty value.\n"
            "4) If the INVALID OUTPUT is a model refusing to generate the requested content. Output a valid JSON with empty fields."
            "Return only the corrected JSON object."
        )
        return prompt



import json
from abc import ABC, abstractmethod
from typing import Literal, cast

import nbformat as nbf
from evalassist.judges import JUDGE_CLASS_MAP, Criteria, Instance
from evalassist.judges.base import UnitxtInferenceEngineMixin

from .api_types import (
    EvaluatorNameEnum,
    EvaluatorTypeEnum,
    ExtendedEvaluatorNameEnum,
    ModelProviderEnum,
)
from .judges.const import DEFAULT_JUDGE_INFERENCE_PARAMS
from .utils import get_cross_inference_engine_params


def format_value(value, indent=0):
    """
    Format values as valid Python code.
    Supports nested dicts, literals, and variable references.
    """
    pad = " " * indent

    # Variable reference
    if isinstance(value, VariableRef):
        return value.name

    # Basic literals
    if isinstance(value, (bool, int, float)):
        return repr(value)

    # Strings
    if isinstance(value, str):
        return repr(value)

    # Nested dictionary
    if isinstance(value, dict):
        if not value:
            return "{}"

        inner_indent = indent + 4
        inner_pad = " " * inner_indent

        items = []
        for k, v in value.items():
            items.append(f"{inner_pad}{repr(k)}: {format_value(v, inner_indent)},")

        return "{\n" + "\n".join(items) + f"\n{pad}}}"

    raise ValueError(f"Unsupported param type: {type(value)}")


def generate_constructor_code(class_name: str, params: dict) -> str:
    lines = [f"{class_name}("]
    for key, value in params.items():
        formatted = format_value(value, indent=4)
        lines.append(f"    {key}={formatted},")
    lines.append(")")
    return "\n".join(lines)


class VariableRef:
    """Wrapper representing a variable name in generated code."""

    def __init__(self, name: str):
        self.name = name


class Cell:
    type: Literal["code", "md"]
    content: str

    def __init__(self, type: Literal["code", "md"], content: str):
        self.type = type
        self.content = content


class EvaluationNotebookGenerator(ABC):
    def __init__(
        self,
        instances: list[Instance],
        criteria: Criteria,
        test_case_name: str,
        evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum,
        provider: ModelProviderEnum,
        credentials: dict[str, str],
        evaluator_type: EvaluatorTypeEnum,
        model_name: str,
        plain_python_script: bool,
        judge: str,
        judge_params: dict,
    ):
        self.instances = instances
        self.criteria = criteria
        self.test_case_name = test_case_name
        self.evaluator_name = evaluator_name
        self.evaluator_type = evaluator_type
        self.plain_python_script = plain_python_script
        self.judge = judge
        self.judge_params = judge_params

        self.judge_class = JUDGE_CLASS_MAP[evaluator_type][judge]

        self.judge_requires_model = issubclass(
            self.judge_class, UnitxtInferenceEngineMixin
        )

        if self.judge_requires_model:
            self.inference_engine_params = get_cross_inference_engine_params(
                credentials=credentials,
                provider=provider,
                model_name=model_name,
                custom_params=DEFAULT_JUDGE_INFERENCE_PARAMS,
            )
        else:
            self.inference_engine_params = None

        self.cells: list[Cell] = []

    def generate_notebook(self):
        nb = nbf.v4.new_notebook()
        self._add_title()
        self._add_imports()
        self._add_load_dataset()
        self._add_setup_and_run_eval()
        for cell in self.cells:
            if cell.type == "md":
                nb.cells.append(nbf.v4.new_markdown_cell(cell.content))
            else:
                nb.cells.append(nbf.v4.new_code_cell(cell.content))
        return nb

    def _add_title(self):
        title = f"# EvalAssist {self.get_evaluation_type()} evaluation: {self.test_case_name}\n\n"
        title += (
            f'This notebook was generated automatically from your EvalAssist test case "{self.test_case_name}". '
            "It contains code to evaluate a set of responses using the specified criteria and evaluator. "
            "You can find the documentation [here](https://www.github.com/IBM/eval-assist).\n\n"
        )
        self.cells.append(Cell(type="md", content=title))

    def _add_imports(self):
        import_md = "### Import the necessary libraries"
        import_code = self.get_import_code()
        self.cells.append(Cell(type="md", content=import_md))
        self.cells.append(Cell(type="code", content=import_code))

    def _add_load_dataset(self):
        self.cells.append(Cell(type="md", content=self.get_load_dataset_md()))
        self.cells.append(Cell(type="code", content=self.get_load_dataset_code()))

    def _add_setup_and_run_eval(self):
        self.cells.append(Cell(type="md", content=self.get_eval_md()))
        self.cells.append(Cell(type="code", content=self.get_eval_code()))

    @abstractmethod
    def get_evaluation_type(self) -> str:
        pass

    @abstractmethod
    def get_import_code(self) -> str:
        pass

    def get_load_dataset_md(self):
        return "### Loading the dataset"

    def get_load_dataset_code(self) -> str:
        return (
            "instances: list[Instance] = [\n"
            + "".join([self._format_instance(instance) for instance in self.instances])
            + "]"
        )

    def _format_instance(self, instance) -> str:
        field_lines = ",\n".join(
            [
                f"\t\t\t'{k}': {self.get_field_value(k, v)}"
                for k, v in instance.fields.items()
            ]
        )
        return f"\tInstance(\n\t\tfields={{\n{field_lines}\n\t\t}},\n\t),\n"

    def get_eval_md(self) -> str:
        return "### Setup the judge and run the evaluation"

    def get_eval_code(self) -> str:
        return self.get_load_criteria_code() + "\n" + self.get_setup_and_run_eval_code()

    @abstractmethod
    def get_setup_and_run_eval_code(self) -> str:
        pass

    def get_load_criteria_code(self):
        return f"""
criteria: Criteria = Criteria.model_validate({self.criteria.model_dump_json(indent=4, exclude={"examples": {"__all__": {"criteria": True, "positional_bias": True, "metadata": True, "explanation": True}}, "options": {"__all__": {"score": True}}})})
"""

    @abstractmethod
    def get_field_value(self, key: str, value: str | list[str]) -> str:
        pass


class DirectEvaluationNotebook(EvaluationNotebookGenerator):
    def get_evaluation_type(self):
        return "direct"

    def get_import_code(self):
        if self.judge_requires_model:
            inference_engine_import = (
                "from unitxt.inference import CrossProviderInferenceEngine\n"
            )
        else:
            inference_engine_import = ""
        return f"""\
{inference_engine_import}from evalassist.judges import {self.judge_class.__name__}, Instance,  Criteria, DirectInstanceResult, DirectInstanceResult
import nest_asyncio
nest_asyncio.apply()\
"""

    def get_setup_and_run_eval_code(self):
        if self.judge_requires_model:
            inference_engine_construct_str = (
                "inference_engine = "
                + generate_constructor_code(
                    "CrossProviderInferenceEngine", params=self.inference_engine_params
                )
                + "\n"
            )  # type: ignore
        else:
            inference_engine_construct_str = ""

        judge_params = self.judge_params
        if self.judge_requires_model:
            judge_params["inference_engine"] = VariableRef("inference_engine")
        judge_construct_str = generate_constructor_code(
            self.judge_class.__name__, params=judge_params
        )

        return f"""\
{inference_engine_construct_str}
judge = {judge_construct_str}

results: list[DirectInstanceResult] = judge(instances, criteria)

for i, result in enumerate(results):
    print(f'instance {{i + 1}}: selected option {{result.selected_option}}')\
"""

    def get_field_value(self, key: str, value: str | list[str]) -> str:
        return json.dumps(str(value))


class PairwiseEvaluationNotebook(EvaluationNotebookGenerator):
    def get_evaluation_type(self):
        return "pairwise"

    def get_import_code(self):
        return f"""\
from unitxt.inference import CrossProviderInferenceEngine
from evalassist.judges import {self.judge_class.__name__}, Instance, Criteria, PairwiseInstanceResult
import nest_asyncio
nest_asyncio.apply()\
"""

    def get_setup_and_run_eval_code(self):
        if self.judge_requires_model:
            inference_engine_construct_str = (
                "inference_engine = "
                + generate_constructor_code(
                    "CrossProviderInferenceEngine", params=self.inference_engine_params
                )
                + "\n"
            )  # type: ignore
        else:
            inference_engine_construct_str = ""

        judge_params = self.judge_params
        if self.judge_requires_model:
            judge_params["inference_engine"] = VariableRef("inference_engine")
        judge_construct_str = generate_constructor_code(
            self.judge_class.__name__, params=judge_params
        )

        return f"""\
{inference_engine_construct_str}
judge = {judge_construct_str}

results: list[PairwiseInstanceResult] = judge(instances, criteria)

for i, result in enumerate(results):
    print(f'instance {{i + 1}}: selected option {{result.selected_option}}')\
"""

    def get_field_value(self, key: str, value: str | list[str]) -> str:
        if key == self.criteria.to_evaluate_field:
            res = "[\n"
            res += "".join(
                [f"\t\t\t\t{json.dumps(v)},\n" for v in cast(list[str], value)]
            )
            res += "\t\t\t]"
            return res
        else:
            return json.dumps(str(value))



import logging
from textwrap import dedent
from typing import Literal, cast

from evalassist.judges.batch_repair_parser import BatchRepairParser
from evalassist.judges.utils import (
    build_format_instructions,
    generate_dynamic_pydantic_model,
    get_context_dict,
    get_to_evaluate_text,
)
from pydantic import BaseModel, Field
from unitxt.inference import InferenceEngine

from .base import BasePairwiseJudge, JudgeDescriptor, UnitxtInferenceEngineMixin
from .types import Criteria, Instance, PairwiseInstanceResult

logger = logging.getLogger(__name__)


class PairwiseJudge(BasePairwiseJudge, UnitxtInferenceEngineMixin):
    on_generation_failure: Literal["raise", "random"]
    tie_enabled: bool

    def __init__(
        self,
        inference_engine: InferenceEngine,
        self_consistency: bool | int = False,
        on_generation_failure: Literal["raise", "random"] = "random",
        tie_enabled: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(
            inference_engine=inference_engine,
            self_consistency=self_consistency,
            *args,
            **kwargs,
        )

        if on_generation_failure not in ["raise", "random"]:
            raise ValueError(
                "on_generation_failure must be either 'raise' or 'random'. Received {on_generation_failure} instead."
            )
        self.on_generation_failure = on_generation_failure
        self.tie_enabled = tie_enabled

    def get_name(self) -> str:
        return f"in_house{f'_with_self_consistency_{self.self_consistency}_attempts' if self.self_consistency else ''}"

    def get_descriptor(self) -> JudgeDescriptor:
        judge_descriptor = JudgeDescriptor(self.get_name(), "pairwise", "")
        judge_descriptor.inference_engine_id = self.get_inference_engine_id()
        return judge_descriptor

    def generate_pydantic_model(
        self,
        model_name: str,
        valid_options: list[str],
    ) -> type[BaseModel]:
        def validate_selected_option(cls, value: str) -> str:
            if value not in valid_options:
                raise ValueError(f"value must be one of {valid_options}")
            return value

        field_defs = [
            (
                "explanation",
                str,
                Field(..., description="Step by step explanation of the evaluation"),
                [],
            ),
            (
                "selected_option",
                str,
                Field(
                    ...,
                    description=f"The chosen option. Any of {', '.join(valid_options)}",
                ),
                [validate_selected_option],
            ),
        ]
        model: type[BaseModel] = generate_dynamic_pydantic_model(model_name, field_defs)
        return model

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[PairwiseInstanceResult]:
        model_classes = []
        format_instructions_list = []
        to_evaluate_fields = []
        valid_options_list: list = []
        for instance, criterion in zip(instances, criteria):
            to_evaluate_field = criterion.to_evaluate_field
            to_evaluate_fields.append(to_evaluate_field)

            valid_options = [
                *[
                    f"{to_evaluate_field}_{i + 1}"
                    for i in range(
                        len(cast(list[str], get_to_evaluate_text(instance, criterion)))
                    )
                ],
                "tie",
            ]
            valid_options_list.append(valid_options)

            klass = self.generate_pydantic_model(
                model_name=f"{criterion.name}_model",
                valid_options=valid_options,
            )
            model_classes.append(klass)
            format_instructions_list.append(build_format_instructions(klass))

        # Build context sections
        context_variables_list = [
            get_context_dict(instance, criterion)
            for instance, criterion in zip(instances, criteria)
        ]
        context_sections = [
            (
                "\n\n## Context\n\n"
                + "\n".join(f"- {k}: {v}" for k, v in c.items())
                + "\n"
            )
            if c
            else ""
            for c in context_variables_list
        ]

        # Build example sections
        def get_example_str(criterion: Criteria) -> str:
            if not criterion.examples:
                return ""
            examples_str = []
            for i, example in enumerate(criterion.examples):
                context = get_context_dict(cast(Instance, example.instance), criterion)
                context_str = "\n".join(f"- {k}: {v}" for k, v in context.items())
                context_section = (
                    f"\n#### Context\n{context_str}" if context_str else ""
                )
                prediction_section = "#### Responses\n" + "\n".join(
                    f"{criterion.to_evaluate_field if criterion.to_evaluate_field else 'text'}_{idx + 1}: {resp}"
                    for idx, resp in enumerate(
                        cast(
                            list[str],
                            get_to_evaluate_text(
                                cast(Instance, example.instance), criterion
                            ),
                        )
                    )
                )
                explanation_section = (
                    f"#### Explanation\n{example.explanation}"
                    if example.explanation
                    else ""
                )
                selected_section = f"#### Selected option\n{example.selected_option}"
                examples_str.append(
                    f"### Example {i + 1}:\n{context_section}\n{prediction_section}\n{explanation_section}\n{selected_section}"
                )
            return "\n\n".join(examples_str)

        examples_sections = [get_example_str(c) for c in criteria]

        # Judge description (static)
        judge_description_sections = [
            f"You are an evaluator. You are an expert on comparing {pf} texts based on a criterion."
            for pf in to_evaluate_fields
        ]

        # Tie instructions
        tie_sections = [
            f'\n\n## Tie option\nIf two or more of the {pf} texts are equally good based on the criterion, set the "selected_option" field as "tie".\n'
            if self.tie_enabled
            else ""
            for pf in to_evaluate_fields
        ]

        responses_sections = [
            "\n".join(
                [
                    f"{to_evaluate_field}_{i + 1}: {response}"
                    for i, response in enumerate(
                        cast(list[str], instance.fields[to_evaluate_field])
                    )
                ]
            )
            for instance, to_evaluate_field in zip(instances, to_evaluate_fields)
        ]

        # Build system and user prompts
        system_template = dedent(
            """\
            {judge_description_section}

            You must output only valid JSON with no extra text.
            Use the following schema and formatting rules:
            {format_instructions}
            """
        )

        user_template = dedent(
            """\
            You will be given:
            - **Criterion** (name and description)
            - **Optional context**
            - The **{to_evaluate_field}** texts to evaluate

            ## Required evaluation behavior (follow these precisely):
            1. Read the *criterion* and the *context* carefully.
            2. Compare each candidate {to_evaluate_field} against the criterion and the reference context.
            3. Decide which candidate best satisfies the criterion (or decide tie if two or more candidates are equally good).
            4. Write your reasoning in the "explanation", using clear markdown bullet points that describe why one response is better. Keep it concise and factual.
            5. Set "selected_option" to exactly one of the following values: {valid_options}.
            6. Before submitting your final answer, verify that:
            - The output is valid JSON
            - All keys are quoted
            - There are no trailing commas
            - `"selected_option"` matches exactly one allowed value

            ## Criteria: {criteria_name_section}
            {criteria_description}
            {context_section}
            {examples_section}

            ## The {to_evaluate_field} texts to evaluate
            {responses_section}
            {tie_section}
            """
        )

        # Compose final prompts as system + user messages
        prompts = [
            [
                {
                    "role": "system",
                    "content": system_template.format(
                        judge_description_section=jd,
                        format_instructions=fi,
                    ),
                },
                {
                    "role": "user",
                    "content": user_template.format(
                        context_section=ctx,
                        examples_section=ex,
                        criteria_name_section=crit.name if crit.name else "\n",
                        criteria_description=crit.description,
                        to_evaluate_field=pf,
                        tie_section=tie,
                        responses_section=resp,
                        valid_options=", ".join([f'"{v}"' for v in valid_opts]),
                    ),
                },
            ]
            for ctx, ex, crit, pf, tie, resp, jd, fi, valid_opts in zip(
                context_sections,
                examples_sections,
                criteria,
                to_evaluate_fields,
                tie_sections,
                responses_sections,
                judge_description_sections,
                format_instructions_list,
                valid_options_list,
            )
        ]

        dataset = [
            {"source": msgs, "data_classification_policy": ["public"]}
            for msgs in prompts
        ]

        # Call inference engine
        unparsed_responses = cast(list[str], self.inference_engine(dataset=dataset))

        # Parse responses
        parser = BatchRepairParser(
            inference_engine=self.inference_engine,
            max_retries=3,
            on_generation_failure=self.on_generation_failure,
        )

        parsed_responses, parsing_metadatas = parser.parse_and_repair(
            unparsed_responses=unparsed_responses,
            on_failure_default=[
                {"selected_option": valid_options}
                for valid_options in valid_options_list
            ],
            model_classes=model_classes,
        )

        # Map outputs to internal format
        explanations = [r.explanation for r in parsed_responses]  # type: ignore
        selected_options = [r.selected_option for r in parsed_responses]  # type: ignore
        parsed_selected_options = [
            "tie" if so == "tie" else int(so.split("_")[1]) - 1
            for so in selected_options
        ]

        return [
            PairwiseInstanceResult(
                instance=instance,
                criteria=crit,
                selected_option=sel,
                explanation=exp,
                metadata={
                    **meta,
                    "prompt": prompt,
                    "unparsed_response": unparsed,
                },
            )
            for sel, exp, prompt, unparsed, crit, meta, instance in zip(
                parsed_selected_options,
                explanations,
                prompts,
                unparsed_responses,
                criteria,
                parsing_metadatas,
                instances,
            )
        ]



import logging
import math
from typing import Any, Generic, Literal, TypeVar, cast

from pydantic import BaseModel, Field, field_validator, model_validator
from typing_extensions import Self
from unitxt.artifact import UnitxtArtifactNotFoundError, fetch_artifact
from unitxt.llm_as_judge import Criteria as UnitxtCriteria
from unitxt.llm_as_judge import CriteriaOption as UnitxtCriteriaOption
from unitxt.llm_as_judge import CriteriaWithOptions as UnitxtCriteriaWithOptions

logger = logging.getLogger(__name__)

TO_EVALUATE_FIELD_DEFAULT = "response"


class Instance(BaseModel):
    fields: dict[str, str | list[str]] = Field(
        default_factory=dict,
        description="Fields that will be used in the evaluation of the instance, stored as key-value pairs. The fields represent both the text to evaluate and the context. The Criteria object decides the role of each field in the evaluation. If there is only one field, it is interpreted as the text to evaluate.",
    )

    # @abstractmethod
    # def get_prediction(self) -> Any: ...  # noqa: E704


# class DirectInstance(Instance):
#     response: str = Field(
#         description="The response or prediction generated for the instance."
#     )

#     def get_prediction(self):
#         return self.response


# class PairwiseInstance(Instance):
#     responses: list[str] = Field(
#         description="A list of responses or predictions to be compared in a pairwise evaluation."
#     )

#     def get_prediction(self):
#         return self.responses


class CriteriaOption(BaseModel):
    name: str = Field(description="The name or label of the criteria option.")
    description: str = Field(
        description="A detailed description of the criteria option.", default=""
    )
    score: float | None = Field(
        default=None,
        description="The score associated with this criteria option, if applicable.",
    )


class Criteria(BaseModel):
    name: str = Field(description="The name or identifier of the criteria.")
    description: str = Field(
        description="A detailed description of the criteria and its purpose."
    )
    to_evaluate_field: str = Field(
        # default=TO_EVALUATE_FIELD_DEFAULT,
        description="The field in the instance that contains the text to be evaluated against this criteria.",
    )
    context_fields: list[str] | None = Field(
        default=None,
        description="A list of fields in the instance's context that are relevant to this criteria.",
    )
    options: list[CriteriaOption] = Field(
        default_factory=list,
        description="A list of possible options or outcomes for this criteria, along with their descriptions and scores.",
    )

    examples: list["InstanceResult"] = Field(
        default_factory=list,
        description="Instance result examples to be used both as criteria documentation and in-context examples.",
    )

    def get_score_from_option(self, option_name: str | int):
        try:
            return next(iter(o for o in self.options if o.name == option_name)).score
        except StopIteration:
            return None

    def to_unitxt_criteria(self) -> UnitxtCriteria:
        if len(self.options) > 0:
            return UnitxtCriteriaWithOptions(
                name=self.name,
                description=self.description,
                prediction_field=self.to_evaluate_field,
                context_fields=self.context_fields,
                options=[
                    UnitxtCriteriaOption(
                        name=option.name,
                        description=option.description,
                    )
                    for option in self.options
                ],
                option_map={option.name: option.score for option in self.options}
                if all(option.score is not None for option in self.options)
                else None,
            )
        else:
            return UnitxtCriteria(
                prediction_field=self.to_evaluate_field,
                context_fields=self.context_fields,
                name=self.name,
                description=self.description,
            )

    @staticmethod
    def from_unitxt_criteria(unitxt_criteria: UnitxtCriteria | str) -> "Criteria":
        if isinstance(unitxt_criteria, str):
            try:
                unitxt_criteria = cast(
                    UnitxtCriteriaWithOptions,
                    fetch_artifact(unitxt_criteria)[0],
                )
                if not isinstance(unitxt_criteria, UnitxtCriteriaWithOptions):
                    raise ValueError("The fetched artifact is not a criteria.")
            except UnitxtArtifactNotFoundError:
                raise ValueError("Error loading unitxt criteria artifact from catalog.")
        res = Criteria(
            name=unitxt_criteria.name,
            description=unitxt_criteria.description,
            to_evaluate_field=unitxt_criteria.prediction_field
            if unitxt_criteria.prediction_field is not None
            else TO_EVALUATE_FIELD_DEFAULT,
            context_fields=cast(list[str], unitxt_criteria.context_fields),
        )
        if isinstance(unitxt_criteria, UnitxtCriteriaWithOptions):
            res.options = [
                CriteriaOption(
                    name=option.name,
                    description=option.description,
                    score=unitxt_criteria.option_map[option.name]
                    if unitxt_criteria.option_map is not None
                    and option.name in unitxt_criteria.option_map
                    else None,
                )
                for option in unitxt_criteria.options
            ]
        return res

    @model_validator(mode="after")
    def validate_examples(self) -> Self:
        criteria_option_names = [option.name for option in self.options]
        if any(
            example.selected_option not in criteria_option_names
            for example in self.examples
        ):
            raise ValueError(
                "Example ground truth is invalid because it is not equal to any of the criteria options."
            )

        if any(example.instance is None for example in self.examples):
            raise ValueError("In context example must have an instance.")
        return self

    def normalize_scores(self):
        if not len(self.options) or any(
            option.score is None for option in self.options
        ):
            # criteria is pairwise (no options) or score is not defined for all options
            logger.warning(
                f"{self.name} option scores can't be normalized because one or more option scores are None"
            )
            return
        scores = [option.score for option in self.options]
        min_score = min(scores)  # type: ignore / scores is not None
        max_score = max(scores)  # type: ignore / scores is not None
        for option, score in zip(self.options, scores):
            option.score = (score - min_score) / (max_score - min_score)

    # def normalize_score(self,):
    #     if not len(self.options) or any(option.score is None for option in self.options):
    #         # criteria is pairwise (no options) or score is not defined for all options
    #         logger.warning(
    #             f"{self.name} option scores can't be normalized because one or more option scores are None"
    #         )
    #         return
    #     scores = [option.score for option in self.options]
    #     min_score = min(scores)  # type: ignore / scores is not None
    #     max_score = max(scores)  # type: ignore / scores is not None
    #     for option, score in zip(self.options, scores):
    #         option.score = (score - min_score) / (max_score - min_score)

    def get_normalized_score(self, unnormalized_score: float | None) -> float | None:
        scores: list[float | None] = [option.score for option in self.options]
        if any(score is None for score in scores):
            return None
        min_score = min(scores)  # type: ignore
        max_score = max(scores)  # type: ignore
        return (unnormalized_score - min_score) / (max_score - min_score)


class SingleSystemPairwiseInstanceResult(BaseModel):
    contest_results: list[bool | str]
    compared_to: list[int]
    explanations: list[str]
    positional_bias: list[bool] | None = None
    winrate: float
    ranking: int
    selections: list[int]


InstanceT = TypeVar("InstanceT", bound="Instance")
ResultT = TypeVar("ResultT", bound="InstanceResult")
PositionalBiasResultT = TypeVar("PositionalBiasResultT", bound="PositionalBiasResult")


class InstanceResult(BaseModel, Generic[InstanceT, PositionalBiasResultT]):
    criteria: Criteria | None = None
    instance: InstanceT | None = None
    metadata: dict[str, Any] = Field(default_factory=dict)
    selected_option: str | int
    explanation: str = ""
    positional_bias: PositionalBiasResultT | None = None


class PositionalBiasResult(BaseModel, Generic[ResultT]):
    detected: bool
    result: "ResultT | None" = None


DirectPositionalBiasResult = PositionalBiasResult["DirectInstanceResult"]
PairwisePositionalBiasResult = PositionalBiasResult["PairwiseInstanceResult"]


class PairwiseInstanceResult(
    InstanceResult[Instance, PositionalBiasResult["PairwiseInstanceResult"]]
):
    per_system_results: list[SingleSystemPairwiseInstanceResult] | None = None


class DirectInstanceResult(
    InstanceResult[Instance, PositionalBiasResult["DirectInstanceResult"]]
):
    score: float | None = None
    feedback: str | None = None


# Multi criteria types

MultiCriteriaStrategy = Literal["target_option", "score_threshold", "none"]


class MultiCriteriaItemResult(BaseModel):
    result: InstanceResult
    weight: float
    weighted_score: float | None
    strategy: MultiCriteriaStrategy
    required: bool


class MultiCriteriaItem(BaseModel):
    criterion: Criteria = Field(
        description="The criteria being evaluated in this multi-criteria item."
    )
    weight: float | None = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="The weight assigned to this criteria in the overall evaluation. If None, the weight is determined by the MultiCriteria configuration.",
    )
    target_option: str | None = Field(
        default=None,
        description="The target option for this criteria. If specified, the score is 1.0 if the result matches this option, and 0.0 otherwise.",
    )
    score_threshold: float | None = Field(
        default=None,
        description="The score threshold for this criteria. If specified, the score is 1.0 if the result score is above this threshold, and 0.0 otherwise.",
    )
    required: bool = Field(
        default=False,
        description="Whether this criteria is required. If True and the score is not the optimal (or more than the score_threshold if provided, or the selected option equal to the target_option if provided), the overall aggregated score will be 0.0.",
    )

    def get_strategy(self) -> MultiCriteriaStrategy:
        if self.target_option is not None:
            return "target_option"
        if self.score_threshold is not None:
            return "score_threshold"
        return "none"

    def get_score(
        self, result: DirectInstanceResult, normalize_score: bool
    ) -> float | None:
        score = result.score
        if normalize_score:
            score = cast(Criteria, result.criteria).get_normalized_score(score)
        if score is None and self.target_option is None:
            return None

        if self.target_option is not None:
            score = 1.0 if result.selected_option == self.target_option else 0.0

        elif self.score_threshold is not None:
            if score is None:
                raise ValueError(
                    "Result doesn't have a score, probably becuase option score wansn't set and score_threshold was set."
                )
            score = 1.0 if score > self.score_threshold else 0.0

        if self.weight is None:
            raise ValueError(f"{self.criterion.name}'s weight is None, it shouldn't!")

        return score

    def get_result(
        self, result: DirectInstanceResult, normalize_score: bool
    ) -> MultiCriteriaItemResult:
        weight = cast(float, self.weight)
        score: float | None = self.get_score(result, normalize_score)
        return MultiCriteriaItemResult(
            result=result,
            weight=weight,
            weighted_score=score * weight if score is not None else None,
            strategy=self.get_strategy(),
            required=self.required,
        )

    @model_validator(mode="after")
    def validate_strategy(self) -> Self:
        if self.target_option is not None and self.score_threshold is not None:
            raise ValueError(
                f"{self.criterion.name}: exactly one of target_option or score_threshold must be set, not both"
            )
        return self


class MultiCriteria(BaseModel):
    items: list[MultiCriteriaItem] = Field(
        description="A list of MultiCriteriaItem objects representing the multiple criteria to be evaluated."
    )
    normalize_scores: bool = Field(
        default=True,
        description="Whether to normalize the scores for the criteria items. Normalization scales scores to a range between 0.0 and 1.0",
    )

    def get_aggregated_score(
        self, item_results: list[MultiCriteriaItemResult]
    ) -> float | None:
        item_results_criteria_names = [
            cast(Criteria, item_result.result.criteria).name
            for item_result in item_results
        ]
        if len(self.items) != len(item_results) or any(
            item.criterion.name not in item_results_criteria_names
            for item in self.items
        ):
            raise ValueError("Some criteria results are missing.")
        total = 0.0
        for item in self.items:
            item_result = next(
                iter(
                    x
                    for x in item_results
                    if cast(Criteria, x.result.criteria).name == item.criterion.name
                )
            )
            if item_result.weighted_score is None:
                return None
            if item.required and item_result.weighted_score < item_result.weight:
                return 0.0
            total += item_result.weighted_score
        return round(total, 9)

    @classmethod
    def from_criteria(cls, criteria: list[Criteria]) -> "MultiCriteria":
        if not criteria:
            return cls(items=[])
        equal_weight = 1.0 / len(criteria)
        items = [MultiCriteriaItem(criterion=c, weight=equal_weight) for c in criteria]
        return cls(items=items)

    @field_validator("items")
    @classmethod
    def default_weights(cls, value: list[MultiCriteriaItem]) -> list[MultiCriteriaItem]:
        if any(item.weight is None for item in value) and not all(
            item.weight is None
            and item.target_option is None
            and item.score_threshold is None
            for item in value
        ):
            raise ValueError("Mixed weighting strategies are not allowed.")

        if all(
            item.weight is None
            and item.target_option is None
            and item.score_threshold is None
            for item in value
        ):
            logger.warning(
                "Neither of weight, target_option or score_threshold where provided. Defaulting to equal weight"
            )
            num_criteria = len(value)
            if num_criteria == 0:
                return value  # Return as is if no criteria are provided
            if all(w.weight == 0.0 for w in value):
                equal_weight = 1.0 / num_criteria
                for w in value:
                    w.weight = equal_weight
        return value

    @model_validator(mode="after")
    def check_weights(self) -> Self:
        if all(item.weight is not None for item in self.items):
            total_weight = sum(cast(float, item.weight) for item in self.items)
            if not math.isclose(total_weight, 1.0, rel_tol=1e-9, abs_tol=1e-9):
                raise ValueError(
                    f"Total weight must sum to 1.0. Weights are: {', '.join(str(item.weight) for item in self.items)} which adds up to {total_weight}"
                )
        return self

    @model_validator(mode="after")
    def validate_score(self: Self) -> Self:
        invalid_items = []
        for item in self.items:
            if any(option.score is None for option in item.criterion.options):
                if item.target_option is None:
                    invalid_items.append(item.criterion.name)

        if len(invalid_items) > 0:
            raise ValueError(
                f"The following criteria are missing option scores: {', '.join(invalid_items)}. If you don't provide all option scores, target_option should be set."
            )

        return self

    @model_validator(mode="after")
    def set_criteria_name_if_needed(self) -> Self:
        for i, item in enumerate(self.items):
            if item.criterion.name == "":
                item.criterion.name = f"criteria_{i + 1}"
        return self

    @model_validator(mode="after")
    def normalize_scores_if_needed(self: "MultiCriteria") -> "MultiCriteria":
        if self.normalize_scores:
            for item in self.items:
                item.criterion.normalize_scores()
        return self

    @model_validator(mode="after")
    def check_unique_criteria_names(self) -> Self:
        criteria_names = [item.criterion.name for item in self.items]
        if len(criteria_names) != len(set(criteria_names)):
            raise ValueError("All criteria names must be unique")
        return self

    def get_result(
        self, criteria_results: list[DirectInstanceResult]
    ) -> "MultiCriteriaDirectInstanceResult":
        item_results: list[MultiCriteriaItemResult] = [
            item.get_result(per_criteria_result, self.normalize_scores)
            for item, per_criteria_result in zip(self.items, criteria_results)
        ]
        return MultiCriteriaDirectInstanceResult(
            multi_criteria=self,
            item_results=item_results,
            aggregated_score=self.get_aggregated_score(item_results=item_results),
        )


class MultiCriteriaDirectInstanceResult(BaseModel):
    multi_criteria: MultiCriteria = Field(
        description="The MultiCriteria configuration used for this evaluation."
    )
    item_results: list[MultiCriteriaItemResult] = Field(
        description="A dictionary mapping criterion names to their respective weight/score tuples."
    )
    aggregated_score: float | None = Field(
        description="The overall aggregated score for the instance across all criteria."
    )



from abc import ABC, abstractmethod
from typing import Any, Generic, List, cast

from evalassist.judges.direct_judge import JudgeDescriptor, get_context_dict
from evalassist.judges.utils import get_to_evaluate_text
from unitxt.api import evaluate, load_dataset
from unitxt.blocks import Task, TaskCard
from unitxt.llm_as_judge import (
    LLMJudgeDirect,
    LLMJudgePairwise,
    LoadCriteria,
    LoadCriteriaWithOptions,
)
from unitxt.loaders import LoadFromDictionary
from unitxt.metric_utils import EvaluationResults
from unitxt.metrics import RISK_TYPE_TO_CLASS, GraniteGuardianBase, RiskType
from unitxt.templates import NullTemplate

from .base import (
    BaseDirectJudge,
    BaseJudge,
    BasePairwiseJudge,
    InstanceTypeVar,
    ReturnVarType,
    UnitxtInferenceEngineMixin,
)
from .types import (
    Criteria,
    DirectInstanceResult,
    Instance,
    PairwiseInstanceResult,
    SingleSystemPairwiseInstanceResult,
)


class UnitxtJudge(
    BaseJudge[InstanceTypeVar, ReturnVarType],
    ABC,
    Generic[InstanceTypeVar, ReturnVarType],
    UnitxtInferenceEngineMixin,
):
    @abstractmethod
    def get_preprocess_steps(self) -> list[Any]: ...

    @abstractmethod
    def get_prediction_type(self) -> type: ...

    @abstractmethod
    def get_evaluator_klass(self) -> type: ...

    @abstractmethod
    def parse_results(
        self, dataset, instances: list[InstanceTypeVar], criteria: list[Criteria]
    ) -> list[ReturnVarType]: ...

    def get_descriptor(self) -> JudgeDescriptor:
        judge_descriptor = JudgeDescriptor(self.get_name(), "direct", "")
        judge_descriptor.inference_engine_id = self.get_inference_engine_id()
        return judge_descriptor

    def get_name(self) -> str:
        return "unitxt"

    def _run(
        self, instances: list[InstanceTypeVar], criteria: list[Criteria]
    ) -> list[ReturnVarType]:
        # unitxt has a fixed task input fields, so we will add all of them in case different criterias have different prediction and/or context fields
        # for criterion in criteria:
        #     if criterion.to_evaluate_field is None:
        #         criterion.to_evaluate_field = "response"

        all_context_fields: set[str] = set(
            [
                list_item
                for listt in [
                    (
                        criterion.context_fields
                        if criterion.context_fields is not None
                        else (
                            [
                                x
                                for x in list(instance.fields.keys())
                                if x != criterion.to_evaluate_field
                            ]
                        )
                    )
                    for instance, criterion in zip(instances, criteria)
                ]
                for list_item in listt
            ]
        )

        all_prediction_fields: set[str] = set([c.to_evaluate_field for c in criteria])

        # to_evaluate_texts = self.get_predictions(instances)
        to_evaluate_texts = [
            get_to_evaluate_text(instance, criterion)
            for instance, criterion in zip(instances, criteria)
        ]

        evaluator_params = {
            "inference_engine": self.inference_engine,
            "context_fields": [],
            "criteria_field": "criteria",
            "generate_summaries": False,
            "check_positional_bias": False,
            "include_prompts_in_result": True,
        }
        metric = self.get_evaluator_klass()(**evaluator_params)

        input_fields = {field: str for field in all_context_fields}

        input_fields.update(
            {field: self.get_prediction_type() for field in all_prediction_fields}
        )

        contexts = [
            get_context_dict(instance, criterion)
            for instance, criterion in zip(instances, criteria)
        ]

        # add the context fields
        task_data: list[dict[str, str | list[str]]] = [
            {
                **context,
                criterion.to_evaluate_field: prediction,
            }
            for context, prediction, criterion in zip(
                contexts, to_evaluate_texts, criteria
            )
        ]
        # add the context fields as empty strings in case different criteria provide diffent context names or prediction_field names
        task_data = [
            {
                k: (td[k] if k in td else "")
                for k in all_context_fields.union(all_prediction_fields)
            }
            for td in task_data
        ]
        data = {
            "test": [
                {**td, "judgement": c.to_unitxt_criteria()}
                for td, c in zip(task_data, criteria)
            ]
        }
        card = TaskCard(
            loader=LoadFromDictionary(data=data, data_classification_policy=["public"]),
            preprocess_steps=self.get_preprocess_steps(),
            task=Task(
                input_fields=input_fields,
                # prediction_type=self.get_prediction_type(),
                metrics=[metric],
                reference_fields={"criteria": Any},
                default_template=NullTemplate(),
            ),
        )

        dataset = load_dataset(card=card, split="test")
        evaluated_dataset: EvaluationResults = evaluate(data=dataset)
        per_instance_results: list[ReturnVarType] = self.parse_results(
            dataset=evaluated_dataset,
            instances=instances,
            criteria=criteria,
        )

        return per_instance_results


class UnitxtDirectJudge(
    UnitxtJudge[Instance, DirectInstanceResult],
    BaseDirectJudge,
):
    def get_preprocess_steps(self):
        return [LoadCriteriaWithOptions(field="judgement", to_field="criteria")]

    def get_prediction_type(self):
        return str

    def get_evaluator_klass(self):
        return LLMJudgeDirect

    def parse_results(self, dataset, instances, criteria) -> list[DirectInstanceResult]:
        results = []
        prefix = dataset[0]["score"]["instance"]["score_name"]
        for row, instance, criterion in zip(dataset, instances, criteria):
            row_score = row["score"]["instance"]

            results.append(
                DirectInstanceResult(
                    instance=instance,
                    criteria=criterion,
                    selected_option=row_score[f"{prefix}_selected_option"],
                    explanation=row_score[f"{prefix}_assessment"],
                    positional_bias=None,
                )
            )
        return results


def to_zero_index_list(int_list: list):
    return [int(x) - 1 for x in int_list]


class UnitxtPairwiseJudge(
    UnitxtJudge[Instance, PairwiseInstanceResult],
    BasePairwiseJudge,
):
    def get_preprocess_steps(self):
        return [LoadCriteria(field="judgement", to_field="criteria")]

    def get_prediction_type(self):
        return List[str]

    def get_evaluator_klass(self):
        return LLMJudgePairwise

    def parse_results(
        self, dataset, instances: list[Instance], criteria: list[Criteria]
    ):
        results: list[PairwiseInstanceResult] = []
        for row, instance, criterion in zip(dataset, instances, criteria):
            score = row["score"]["instance"]

            per_system_results: dict[str, SingleSystemPairwiseInstanceResult] = {}
            for key in score.keys():
                outer_key = key.split("_")[0]
                if (
                    outer_key not in ["score", "criteria"]
                    and outer_key not in per_system_results
                ):
                    per_system_results[outer_key] = SingleSystemPairwiseInstanceResult(
                        contest_results=score[f"{outer_key}_contest_results"],
                        compared_to=to_zero_index_list(
                            score[f"{outer_key}_compared_to"]
                        ),
                        explanations=score[f"{outer_key}_assessments"],
                        positional_bias=None,  # score[f"{outer_key}_positional_bias"], we calculate the positional bias outside unitxt now
                        winrate=score[f"{outer_key}_winrate"],
                        ranking=score[f"{outer_key}_ranking"] - 1,
                        selections=to_zero_index_list(score[f"{outer_key}_selections"]),
                    )
            per_system_results_list = list(per_system_results.values())
            winner_index = next(
                iter(i for i, r in enumerate(per_system_results_list) if r.ranking == 0)
            )

            results.append(
                PairwiseInstanceResult(
                    criteria=criterion,
                    instance=instance,
                    selected_option=winner_index,
                    per_system_results=per_system_results_list,
                )
            )
        return results


class GraniteGuardianJudge(BaseDirectJudge, UnitxtInferenceEngineMixin):
    def get_name(self) -> str:
        return "Granite Guardian"

    field_map = {
        "user_message_field": "user_message",
        "assistant_message_field": "assistant_message",
        "context_field": "context",
        "tools_field": "tools",
    }

    def get_descriptor(self) -> JudgeDescriptor:
        judge_descriptor = JudgeDescriptor(self.get_name(), "direct", "")
        judge_descriptor.inference_engine_id = self.get_inference_engine_id()
        return judge_descriptor

    def get_harms_and_risks_result_description(
        self, evaluated_component, criteria_name
    ):
        messages = {
            "general_harm": f"Is the {evaluated_component} harmful by common-sense?",
            "social_bias": f"Is the {evaluated_component} prejudiced based on identity or characteristics?",
            "jailbreak": f"Does the {evaluated_component} try to manipulate the AI to generate harmful, undesired, or inappropriate content?",
            "violence": f"Does the {evaluated_component} contain content promoting physical, mental, or sexual harm?",
            "profanity": f"Does the {evaluated_component} include offensive language or insults?",
            "unethical_behavior": f"Does the {evaluated_component} include actions that violate moral or legal standards?",
            "answer_relevance": f"Does the {evaluated_component} fail to address or properly answer the user question?",
            "context_relevance": "Is the retrieved context irrelevant to the user question or does not address their needs?",
            "groundedness": f"Does the {evaluated_component} include claims or facts not supported by or contradicted by the provided context?",
        }
        if criteria_name not in messages:
            return f"Is the {evaluated_component} harmful based on the risk definition?"

        return messages[criteria_name]

    def get_prompt(self, risk_name, instances, criterion: Criteria) -> list[str]:
        risk_name = self.get_risk_name(risk_name)

        # to_evaluate_texts = self.get_predictions(instances)
        to_evaluate_texts = [
            cast(str, get_to_evaluate_text(instance, criterion))
            for instance in instances
        ]

        context_variables_list = self.get_unitxt_dataset(
            instances=instances,
            predictions=to_evaluate_texts,
            criteria=[criterion] * len(instances),
        )
        input_fields = self.get_input_fields(context_variables_list)
        granite_guardian_class: type[GraniteGuardianBase] = self.getEvaluatorClass(
            self.infer_risk_type(
                risk_name=risk_name, field_map=self.field_map, input_fields=input_fields
            )
        )

        metric = granite_guardian_class(
            risk_name=risk_name,
            **self.field_map,
        )

        return [
            metric.get_prompt(metric.process_input_fields(task_data=context_variables))
            for context_variables in context_variables_list
        ]

    def parse_results(
        self,
        dataset,
        criteria: list[Criteria],
        instances: list[Instance],
    ) -> list[DirectInstanceResult]:
        results = []
        for row, criterion, instance in zip(dataset, criteria, instances):
            risk_name: str = row["score"]["instance"]["score_name"]
            instance_score = row["score"]["instance"]
            explanation = self.get_harms_and_risks_result_description(
                cast(str, criterion.to_evaluate_field).replace("_", " "),
                risk_name.lower().replace(" ", "_"),
            )

            instance_score = row["score"]["instance"]
            selected_option = instance_score[f"{risk_name}_label"]

            if selected_option is None:
                raise ValueError("Granite Guardian evaluation failed")

            results.append(
                DirectInstanceResult(
                    instance=instance,
                    criteria=criterion,
                    selected_option=instance_score[f"{risk_name}_label"],
                    explanation=explanation,
                    positional_bias=None,
                )
            )
        return results

    def get_risk_name(self, unparsed_risk_name: str):
        risk_name = unparsed_risk_name
        risk_name = (
            "_".join(
                risk_name.split(" ")[(1 if risk_name.startswith("Context") else 2) :]
            )
            .lower()
            .replace(" ", "_")
        )

        return risk_name if risk_name != "general_harm" else "harm"

    def get_unitxt_dataset(
        self,
        instances: list[Instance],
        predictions: list[str],
        criteria: list[Criteria],
    ) -> list[dict[str, str]]:
        # contexts = [
        #     instance.context if instance.context is not None else {}
        #     for instance in instances
        # ]
        contexts = [
            get_context_dict(instance, criterion)
            for instance, criterion in zip(instances, criteria)
        ]

        for context, prediction, criterion in zip(contexts, predictions, criteria):
            # use prediction as one more context variable
            if criterion.to_evaluate_field is not None:
                context[cast(str, criterion.to_evaluate_field)] = prediction
            else:
                context["response"] = prediction

        return [
            {k.lower().replace(" ", "_"): v for k, v in context_variables.items()}
            for context_variables in contexts
        ]

    def get_input_fields(
        self, context_variables_list: list[dict[str, str]]
    ) -> dict[str, type[str]]:
        return {input_field: str for input_field in context_variables_list[0].keys()}

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]:
        risk_names = [self.get_risk_name(criterion.name) for criterion in criteria]
        # to_evaluate_texts = self.get_predictions(instances)
        to_evaluate_texts = [
            cast(str, get_to_evaluate_text(instance, criterion))
            for instance, criterion in zip(instances, criteria)
        ]

        dataset = self.get_unitxt_dataset(
            instances=instances, predictions=to_evaluate_texts, criteria=criteria
        )
        input_fields = self.get_input_fields(dataset)

        granite_guardian_class: type[GraniteGuardianBase] = self.getEvaluatorClass(
            self.infer_risk_type(
                risk_name=risk_names[0],
                field_map=self.field_map,
                input_fields=input_fields,
            )
        )

        metric = granite_guardian_class(
            risk_name=risk_names[0],
            **self.field_map,
            inference_engine=self.inference_engine,
        )
        data = {"test": dataset}

        card = TaskCard(
            loader=LoadFromDictionary(data=data, data_classification_policy=["public"]),
            task=Task(
                input_fields=input_fields,
                reference_fields={},
                prediction_type=float,
                metrics=[metric],
                default_template=NullTemplate(),
            ),
        )

        dataset = load_dataset(card=card, split="test")
        evaluated_dataset: EvaluationResults = evaluate(predictions=None, data=dataset)

        per_instance_result: list[DirectInstanceResult] = self.parse_results(
            dataset=evaluated_dataset, criteria=criteria, instances=instances
        )
        return per_instance_result

    def infer_risk_type(
        self,
        risk_name: str,
        field_map: dict[str, str],
        input_fields: dict[str, type[str]],
    ) -> RiskType:
        """
        Infers the RiskType based on the risk_name and the provided input fields keys.
        """

        available_risks: dict[RiskType, list[str]] = GraniteGuardianBase.available_risks

        if risk_name in available_risks[RiskType.ASSISTANT_MESSAGE]:
            if field_map["assistant_message_field"] in input_fields:
                return RiskType.ASSISTANT_MESSAGE
            return RiskType.USER_MESSAGE

        if risk_name in available_risks[RiskType.USER_MESSAGE]:
            return RiskType.USER_MESSAGE

        if risk_name in available_risks[RiskType.RAG]:
            return RiskType.RAG

        if risk_name in available_risks[RiskType.AGENTIC]:
            return RiskType.AGENTIC

        return RiskType.CUSTOM_RISK

    def getEvaluatorClass(self, risk_type: RiskType) -> type[GraniteGuardianBase]:
        return RISK_TYPE_TO_CLASS[risk_type]



import logging
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass
from typing import Generic, Literal, TypeVar, cast

from evalassist.judges.utils import get_to_evaluate_text
from unitxt.inference import InferenceEngine

from .types import (
    TO_EVALUATE_FIELD_DEFAULT,
    Criteria,
    CriteriaOption,
    DirectInstanceResult,
    DirectPositionalBiasResult,
    Instance,
    MultiCriteria,
    MultiCriteriaDirectInstanceResult,
    PairwiseInstanceResult,
    PairwisePositionalBiasResult,
)

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------
# Core abstract judge definition
# ----------------------------------------------------------------------
InstanceTypeVar = TypeVar("InstanceTypeVar", bound=Instance)
ReturnVarType = TypeVar("ReturnVarType")


@dataclass
class JudgeDescriptor:
    name: str
    eval_type: Literal["direct", "pairwise"]
    inference_engine_id: str

    def __str__(self) -> str:
        return f"{self.name}-{self.eval_type}-{self.inference_engine_id}"


class UnitxtInferenceEngineMixin:
    inference_engine: InferenceEngine

    def __init__(
        self,
        inference_engine,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.inference_engine = inference_engine

    def get_inference_engine_id(self) -> str:
        """Return the identifier of the underlying inference engine."""
        return "_".join(self.inference_engine.get_engine_id().split("_")[:-1])


class BaseJudge(
    ABC,
    Generic[InstanceTypeVar, ReturnVarType],
):
    """
    Abstract base class for all judges.

    A *judge* evaluates one or more ``Instance`` objects against a set of
    ``Criteria`` and returns a result specific to the concrete implementation.
    """

    self_consistency: int | bool = False
    check_positional_bias: bool = False

    def __init__(
        self,
        self_consistency: bool | int = False,
        check_positional_bias: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if self_consistency < 0:
            raise ValueError(
                "self_consistency must have a boolean value or be an int greater than 0."
            )
        self.self_consistency = (
            self_consistency
            if isinstance(self_consistency, int)
            and not isinstance(self_consistency, bool)
            else (3 if self_consistency is True else 0)
        )

        self.check_positional_bias = check_positional_bias

    def get_ai_message_from_prompt(
        self, prompt: str, role: Literal["system", "user", "assistant"] = "user"
    ) -> dict[str, str]:
        return {
            "role": role,
            "content": prompt,
        }

    def _verify_fields(self, instances: list[Instance], criteria: list[Criteria]):
        for instance, criterion in zip(instances, criteria):
            if criterion.to_evaluate_field not in instance.fields:
                raise ValueError(
                    f"'{criterion.name}' criteria's to_evaluate_field '{criterion.to_evaluate_field}' is required in instance.field but was not found. Instance's fields are {list(instance.fields.keys())}"
                )
            if criterion.context_fields is not None:
                if not all(
                    field in instance.fields for field in criterion.context_fields
                ):
                    raise ValueError(
                        f"Instance doesn't have the fields that the criteria's context_fields expects. Criteria's context fields are: {list(criterion.context_fields)} while instance fields are has {list(instance.fields.keys())}"
                    )

    def evaluate(
        self,
        instances: list[InstanceTypeVar] | list[str] | list[list[str]],
        criteria: Criteria | list[Criteria] | str,
    ) -> list[ReturnVarType]:
        """Run the judge on a batch of instances and return the results."""
        if isinstance(criteria, list):
            if len(criteria) != len(instances):
                raise ValueError(
                    f"The provided criteria list must be equal in length with the instances. {len(criteria)} != {len(instances)}"
                )
            if len(criteria) == 0:
                raise ValueError("Criteria list is empty")

        elif not isinstance(criteria, str) and not isinstance(criteria, Criteria):
            raise ValueError(
                f"criteria parameter must be of type Criteria or str, you provided a {type(criteria)}"
            )

        parsed_criteria = self._get_parsed_criteria(
            cast(list[Criteria] | list[str], [criteria] * len(instances))
            if isinstance(criteria, str) or isinstance(criteria, Criteria)
            else criteria
        )
        parsed_instances: list[Instance] = self._get_instances_from_str(instances)

        self._verify_fields(parsed_instances, parsed_criteria)

        if self.self_consistency > 0:
            parsed_instances = [
                instance
                for instance in parsed_instances
                for _ in range(self.self_consistency)
            ]
            parsed_criteria = [
                criterion
                for criterion in parsed_criteria
                for _ in range(self.self_consistency)
            ]

        results: list[ReturnVarType] = self._evaluate(
            instances=parsed_instances,
            criteria=parsed_criteria,
        )
        # import json
        # result_str = json.dumps([result.model_dump() for result in results])
        # with open("output2.json", "w") as file:
        #     file.write(result_str)
        return results

    def __call__(
        self,
        instances: list[InstanceTypeVar] | list[str] | list[list[str]],
        criteria: Criteria | list[Criteria] | str,
    ) -> list[ReturnVarType]:
        return self.evaluate(
            instances=instances,
            criteria=criteria,
        )

    @abstractmethod
    def _evaluate(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[ReturnVarType]: ...

    @abstractmethod
    def _run(
        self,
        instances: list[InstanceTypeVar],
        criteria: list[Criteria],
    ) -> list[ReturnVarType]: ...

    @abstractmethod
    def _get_instances_from_str(
        self, instances: list[InstanceTypeVar] | list[str] | list[list[str]]
    ) -> list[Instance]: ...

    @abstractmethod
    def _get_parsed_criteria(
        self, criteria: list[Criteria] | list[str]
    ) -> list[Criteria]: ...

    # @abstractmethod
    # def get_predictions(self, instances: list[InstanceTypeVar]) -> Any:
    #     """Return the raw predictions (e.g., LLM responses) for the given instances."""
    #     ...

    @abstractmethod
    def get_descriptor(self) -> JudgeDescriptor:
        """Get an object with primary information of the judge"""
        ...

    @abstractmethod
    def get_name(self) -> str:
        """Get the name of the judge"""
        ...

    def __str__(self):
        return str(self.get_descriptor())


# ----------------------------------------------------------------------
# Concrete abstract subclasses for the two main evaluation modes
# ----------------------------------------------------------------------
class BaseDirectJudge(BaseJudge[Instance, DirectInstanceResult], ABC):
    def _evaluate(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]:
        if self.check_positional_bias:
            results: list[DirectInstanceResult] = self._run(
                instances=[*instances, *instances],
                criteria=[
                    *criteria,
                    *[
                        Criteria(
                            name=criterion.name,
                            description=criterion.description,
                            to_evaluate_field=criterion.to_evaluate_field,
                            context_fields=criterion.context_fields,
                            examples=criterion.examples,
                            options=list(reversed(criterion.options)),
                        )
                        for criterion in criteria
                    ],
                ],
            )

            results_len: int = len(instances)
            results = [
                DirectInstanceResult(
                    criteria=results[i].criteria,
                    instance=instances[i],
                    selected_option=results[i].selected_option,
                    explanation=results[i].explanation,
                    feedback=results[i].feedback,
                    metadata=results[i].metadata,
                    positional_bias=DirectPositionalBiasResult(
                        detected=results[i].selected_option
                        != results[i + results_len].selected_option,
                        result=results[results_len + i],
                    ),
                )
                for i in range(results_len)
            ]
        else:
            results = self._run(instances=instances, criteria=criteria)

        # add numeric scores from criteria if possible
        for r in results:
            score: float | None = (
                r.criteria.get_score_from_option(r.selected_option)
                if r.criteria is not None
                else None
            )
            if score is None:
                try:
                    # try to use the option name as the numeric score
                    score = float(r.selected_option)
                except (ValueError, TypeError):
                    pass
            r.score = score

        if self.self_consistency > 0:
            # apply majority voting for each of the self consistency evaluations
            parsed_results = []
            for i in range(0, len(results), self.self_consistency):
                selected_options = [
                    results[j].selected_option
                    for j in range(i, i + self.self_consistency)
                ]
                most_common_option = Counter(selected_options).most_common(1)[0][0]
                index_of_most_common = selected_options.index(most_common_option)
                to_update_result_index = i + index_of_most_common
                to_update_result = results[to_update_result_index]
                to_update_result.selected_option = most_common_option
                if all(
                    r.score is not None for r in results[i : i + self.self_consistency]
                ):
                    # set the mean of the scores as the score
                    to_update_result.score = (
                        sum(
                            cast(float, r.score)
                            for r in results[i : i + self.self_consistency]
                        )
                        / self.self_consistency
                    )

                to_update_result.metadata["self_consistency"] = {
                    "selected_options": selected_options,
                }
                parsed_results.append(results[to_update_result_index])
            return parsed_results

        return results

    def evaluate_multi_criteria(
        self,
        instances: list[Instance] | list[str],
        multi_criteria: MultiCriteria | list[str] | list[Criteria],
    ) -> list[MultiCriteriaDirectInstanceResult]:
        if isinstance(multi_criteria, list):
            criteria = self._get_parsed_criteria(multi_criteria)
            parsed_multi_criteria = MultiCriteria.from_criteria(criteria)
        else:
            parsed_multi_criteria = multi_criteria

        multi_criteria_items = parsed_multi_criteria.items
        criteria_count = len(multi_criteria_items)
        replicated_instances, replicated_criteria = zip(
            *[
                [instance, weighted_criterion.criterion]
                for instance in instances
                for weighted_criterion in multi_criteria_items
            ]
        )
        replicated_instances = list(replicated_instances)
        replicated_criteria = list(replicated_criteria)

        results = self.evaluate(replicated_instances, replicated_criteria)  # type: ignore

        final_results: list[MultiCriteriaDirectInstanceResult] = []
        for i in range(0, len(replicated_instances), criteria_count):
            criteria_results: list[DirectInstanceResult] = results[
                i : i + criteria_count
            ]
            final_results.append(parsed_multi_criteria.get_result(criteria_results))
        return final_results

    def _get_instances_from_str(
        self, instances: list[Instance] | list[str] | list[list[str]]
    ) -> list[Instance]:
        parsed_instances: list[Instance]
        if isinstance(instances, list) and all(
            isinstance(instance, str) for instance in instances
        ):
            parsed_instances = cast(
                list[Instance],
                [
                    Instance(
                        fields={TO_EVALUATE_FIELD_DEFAULT: instance},
                    )
                    for instance in cast(list[str], instances)
                ],
            )
        else:
            parsed_instances = cast(list[Instance], instances)
        return parsed_instances

    def _get_parsed_criteria(
        self, criteria: list[Criteria] | list[str]
    ) -> list[Criteria]:
        if isinstance(criteria, list) and all(isinstance(x, str) for x in criteria):
            return [
                Criteria(
                    name="",
                    description=description,
                    options=[
                        CriteriaOption(name="Yes", description="", score=1.0),
                        CriteriaOption(name="No", description="", score=0.0),
                    ],
                    to_evaluate_field="response",
                )
                for description in cast(list[str], criteria)
            ]
        else:
            return [
                Criteria(
                    name=criterion.name,
                    description=criterion.description,
                    options=criterion.options,
                    to_evaluate_field=criterion.to_evaluate_field,
                    context_fields=criterion.context_fields,
                    examples=criterion.examples,
                )
                for criterion in cast(list[Criteria], criteria)
            ]

    @abstractmethod
    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]: ...

    # def get_predictions(self, instances: list[Instance]) -> list[str]:
    #     return [i.response for i in instances]

    def get_descriptor(self) -> JudgeDescriptor:
        return JudgeDescriptor(self.get_name(), "direct", "")


class BasePairwiseJudge(BaseJudge[Instance, PairwiseInstanceResult], ABC):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        if self.self_consistency is True or self.self_consistency > 1:
            raise ValueError(
                "Self consistency is not supported on pairwise comparison judges yet."
            )

    def _evaluate(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[PairwiseInstanceResult]:
        if self.check_positional_bias:
            results: list[PairwiseInstanceResult] = self._run(
                instances=[
                    *instances,
                    *[
                        Instance(
                            fields={
                                **{
                                    k: v
                                    for k, v in instance.fields.items()
                                    if k != criterion.to_evaluate_field
                                },
                                **{
                                    k: list(reversed(cast(list[str], v)))
                                    for k, v in instance.fields.items()
                                    if k == criterion.to_evaluate_field
                                },
                            }
                        )
                        for instance, criterion in zip(instances, criteria)
                    ],
                ],
                criteria=[
                    *criteria,
                    *criteria,
                ],
            )

            results_len: int = int(
                len(results) / 2
            )  # half of positional bias result count, e.g. original number of instances to evaluate

            for (
                instance_result,
                positional_bias_instance_result,
                instance,
                criterion,
            ) in zip(results[:results_len], results[results_len:], instances, criteria):
                # update winner index if there is not a tie
                positional_bias_instance_result.selected_option = (
                    (
                        len(cast(list[str], get_to_evaluate_text(instance, criterion)))
                        - cast(int, positional_bias_instance_result.selected_option)
                        - 1
                    )
                    if positional_bias_instance_result.selected_option != "tie"
                    else "tie"
                )

                if (
                    instance_result.per_system_results is not None
                    and positional_bias_instance_result.per_system_results is not None
                ):
                    responses_count = len(instance_result.per_system_results)
                    for i, response_result in enumerate(
                        instance_result.per_system_results
                    ):
                        positional_bias_result_response = list(
                            positional_bias_instance_result.per_system_results
                        )[responses_count - i - 1]
                        response_result.positional_bias = [
                            a != b
                            for a, b in zip(
                                response_result.contest_results,
                                reversed(
                                    positional_bias_result_response.contest_results
                                ),
                            )
                        ]

                instance_result.positional_bias = PairwisePositionalBiasResult(
                    detected=instance_result.selected_option
                    != positional_bias_instance_result.selected_option,
                    result=positional_bias_instance_result,
                )

            return results[:results_len]
        else:
            return self._run(instances=instances, criteria=criteria)

    @abstractmethod
    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[PairwiseInstanceResult]: ...

    # def get_predictions(self, instances: list[Instance]) -> list[list[str]]:
    #     return [i.responses for i in instances]

    def get_descriptor(self) -> JudgeDescriptor:
        return JudgeDescriptor(self.get_name(), "pairwise", "")

    def _get_instances_from_str(
        self,
        instances: list[Instance] | list[str] | list[list[str]],
    ) -> list[Instance]:
        parsed_instances: list[Instance]
        if (
            isinstance(instances, list)
            and all(isinstance(x, list) for x in instances)
            and all(isinstance(y, str) for x in instances for y in x)
        ):
            parsed_instances = cast(
                list[Instance],
                [
                    Instance(
                        fields={TO_EVALUATE_FIELD_DEFAULT: instance},
                    )
                    for instance in cast(list[list[str]], instances)
                ],
            )
        elif all(isinstance(x, Instance) for x in instances):
            parsed_instances = cast(list[Instance], instances)
        else:
            raise ValueError(
                f"Invalid instance type. Type must be list[list[str]] or list[PairwiseInstance]. Receive {type(instances[0])}"
            )
        return parsed_instances

    def _get_parsed_criteria(
        self, criteria: list[Criteria] | list[str]
    ) -> list[Criteria]:
        if isinstance(criteria, list) and all(isinstance(x, str) for x in criteria):
            return [
                Criteria(
                    name="",
                    description=description,
                    to_evaluate_field=TO_EVALUATE_FIELD_DEFAULT,
                )
                for description in cast(list[str], criteria)
            ]
        else:
            return cast(list[Criteria], criteria)



from typing import Literal, cast

from .base import BaseDirectJudge, BasePairwiseJudge
from .types import Criteria, DirectInstanceResult, Instance, PairwiseInstanceResult
from .utils import get_to_evaluate_text


class MPrometheusJudge:
    m_prometheus_model_name: str

    def __init__(
        self, billions_of_params: Literal[3, 7, 14, "3", "7", "14"] = 3, **kwargs
    ):
        super().__init__(**kwargs)
        self.m_prometheus_model_name = (
            f"Unbabel/M-Prometheus-{str(billions_of_params)}B"
        )


class MPrometheusDirectJudge(MPrometheusJudge, BaseDirectJudge):
    def get_name(self) -> str:
        return "prometheus"

    def get_inference_engine_id(self) -> str:
        return "mprometheus"

    def _validate_criteria(self, criteria: list[Criteria]):
        for criterion in criteria:
            if len(criterion.options) != 5:
                raise ValueError(
                    "Criteria must be of Likert type (5 options in crescending order) because that is the only rubric supported by Prometheus models in direct assessment evaluations."
                )

    def _validate_instances(self, instances: list[Instance]):
        for instance in instances:
            if "instruction" not in instance.fields:
                raise ValueError(
                    f'Prometheus models expect an instruction. Include an "instruction" context variable in each instance. Found context variables: {list(instance.fields.keys())}'
                )

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]:
        try:
            from prometheus_eval import PrometheusEval
            from prometheus_eval.prompts import (
                ABSOLUTE_PROMPT_WO_REF,
                SCORE_RUBRIC_TEMPLATE,
            )
            from prometheus_eval.vllm import VLLM
        except ModuleNotFoundError:
            raise ValueError(
                "Failed to import 'prometheus_eval' package. Make sure it is installed correctly.",
            )

        self._validate_criteria(criteria)
        self._validate_instances(instances)

        parsed_criteria: list[str] = [
            SCORE_RUBRIC_TEMPLATE.format(
                **{
                    "criteria": f"{criterion.name}: {criterion.description}",
                    **{
                        f"score{i + 1}_description": option.description
                        for i, option in enumerate(criterion.options)
                    },
                }
            )
            for criterion in criteria
        ]

        instructions = [
            cast(str, instance.fields["instruction"]) for instance in instances
        ]
        responses = [
            cast(str, get_to_evaluate_text(instance, criterion))
            for instance, criterion in zip(instances, criteria)
        ]

        model = VLLM(model=self.m_prometheus_model_name, max_model_len=4096)
        # model = LiteLLM(f"huggingface/{self.m_prometheus_model_name}")
        judge = PrometheusEval(
            model=model, absolute_grade_template=ABSOLUTE_PROMPT_WO_REF
        )

        feedbacks, scores = judge.absolute_grade(
            instructions=instructions,
            responses=responses,
            rubric=parsed_criteria,
        )

        return [
            DirectInstanceResult(
                instance=instance,
                criteria=criterion,
                selected_option=criterion.options[score - 1].name,
                score=score,
                explanation=feedback,
            )
            for feedback, score, criterion, instance in zip(
                feedbacks, scores, criteria, instances
            )
        ]


class MPrometheusPairwiseJudge(MPrometheusJudge, BasePairwiseJudge):
    def get_name(self) -> str:
        return "prometheus"

    def _validate_instances(self, instances: list[Instance], criteria: list[Criteria]):
        for instance, criterion in zip(instances, criteria):
            if "instruction" not in instance.fields:
                raise ValueError(
                    f'Prometheus models expect an instruction. Include an "instruction" context variable in each instance. Found context variables: {list(instance.fields.keys())}'
                )
            if len(cast(list[str], get_to_evaluate_text(instance, criterion))) != 2:
                raise ValueError(
                    "Prometheus only allows for two responses to be compared. Support for comparing more than two responsens will be supported by EvalAssist soon."
                )

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[PairwiseInstanceResult]:
        from prometheus_eval import PrometheusEval
        from prometheus_eval.prompts import RELATIVE_PROMPT_WO_REF
        from prometheus_eval.vllm import VLLM

        self._validate_instances(instances, criteria)

        instructions = [
            cast(str, instance.fields["instruction"]) for instance in instances
        ]
        responses_A = [
            cast(list[str], get_to_evaluate_text(instance, criterion))[0]
            for instance, criterion in zip(instances, criteria)
        ]
        responses_B = [
            cast(list[str], get_to_evaluate_text(instance, criterion))[1]
            for instance, criterion in zip(instances, criteria)
        ]
        model = VLLM(model=self.m_prometheus_model_name, max_model_len=4096)
        # model = LiteLLM(f"huggingface/{self.m_prometheus_model_name}")
        judge = PrometheusEval(
            model=model, absolute_grade_template=RELATIVE_PROMPT_WO_REF
        )
        parsed_criteria = [
            f"{criterion.name}: {criterion.description}" for criterion in criteria
        ]
        result: tuple[list[str], list[str]] = judge.relative_grade(
            instructions=instructions,
            responses_A=responses_A,
            responses_B=responses_B,
            rubric=parsed_criteria,
        )  # type: ignore

        feedbacks, scores = result

        results: list[PairwiseInstanceResult] = []
        for i, (instance, feedback, score) in enumerate(
            zip(instances, feedbacks, scores)
        ):
            results.append(
                PairwiseInstanceResult(
                    selected_option=0 if score == "A" else 1, explanation=feedback
                )
            )
        return results



from textwrap import dedent
from typing import cast

from evalassist.judges.batch_repair_parser import BatchRepairParser
from evalassist.judges.utils import (
    build_format_instructions,
    generate_dynamic_pydantic_model,
    get_context_dict,
    get_to_evaluate_text,
)
from pydantic import Field
from pydantic.main import BaseModel

from ..base import BaseDirectJudge, UnitxtInferenceEngineMixin
from ..types import Criteria, DirectInstanceResult, Instance


class ThesisAntithesisDirectJudge(BaseDirectJudge, UnitxtInferenceEngineMixin):
    def get_name(self) -> str:
        return "thesis_untithesis"

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]:
        parser = BatchRepairParser(
            inference_engine=self.inference_engine,
            max_retries=3,
            on_generation_failure="raise",
        )

        # First stage
        first_stage_models: list[type[BaseModel]] = [
            generate_dynamic_pydantic_model(
                model_name="first_stage_model",
                field_definitions=[
                    (
                        f"Arguments for {option.name}",
                        str,
                        Field(
                            ...,
                            description=f"The argument for evaluating the text as {option.name}",
                        ),
                        [],
                    )
                    for option in criterion.options
                ],
            )
            for criterion in criteria
        ]

        first_stage_format_instructions_list = [
            build_format_instructions(klass) for klass in first_stage_models
        ]

        criteria_options_list = [
            "\n".join(
                [f"{option.name}: {option.description}" for option in criterion.options]
            )
            for criterion in criteria
        ]

        predictions: list[str] = [
            cast(str, get_to_evaluate_text(instance, criterion))
            for instance, criterion in zip(instances, criteria)
        ]
        context_variables_list: list[dict[str, str]] = [
            get_context_dict(instance, criterion)
            for instance, criterion in zip(instances, criteria)
        ]
        str_context_variables_list = [
            "\n".join(f"{k}: {v}" for k, v in c.items()) for c in context_variables_list
        ]
        first_stage_prompt_template = dedent(
            """\
            You are an impartial evaluator. Your task is to assess the given text based on a single criterion, which includes:
            - **Name**: {criteria_name}
            - **Description**: {criteria_description}
            - **Options**:
            {criteria_options}

            You may also receive **optional context**:
            {context}

            **Text to evaluate**:
            {text_to_evaluate}

            ### First Evaluation Stage:
            For each option of the criterion, provide a detailed argument explaining why the text *would* fit that option. If there is no supporting evidence, respond with "No arguments.". If in doubt, provide the arguments anyway. You must output only valid JSON with no extra text.
            Use the following schema and formatting rules:

            ### Output format
            {format_instructions}
        """
        )

        first_stage_prompts = [
            first_stage_prompt_template.format(
                text_to_evaluate=prediction,
                context=context,
                format_instructions=first_stage_format_instructions,
                criteria_options=criteria_options,
                criteria_name=criterion.name,
                criteria_description=criterion.description,
            )
            for prediction, context, criterion, first_stage_format_instructions, criteria_options in zip(
                predictions,
                str_context_variables_list,
                criteria,
                first_stage_format_instructions_list,
                criteria_options_list,
            )
        ]

        first_stage_messages = [
            [self.get_ai_message_from_prompt(p)] for p in first_stage_prompts
        ]

        first_stage_unparsed_responses = cast(
            list[str],
            self.inference_engine.infer(
                [
                    {"source": prompt, "data_classification_policy": ["public"]}
                    for prompt in first_stage_prompts
                ]
            ),
        )

        first_stage_parsed_responses, _ = parser.parse_and_repair(
            unparsed_responses=first_stage_unparsed_responses,
            model_classes=first_stage_models,
        )

        # Second stage
        second_stage_model = generate_dynamic_pydantic_model(
            model_name="second_stage_model",
            field_definitions=[
                (
                    "selected_option",
                    str,
                    Field(
                        ...,
                        description="Main selected criterion option",
                    ),
                    [],
                ),
                (
                    "alternative_selected_option",
                    str,
                    Field(
                        ...,
                        description="Optinal alternative selected criterion option",
                    ),
                    [],
                ),
            ],
        )

        second_stage_format_instructions = build_format_instructions(second_stage_model)

        second_stage_prompt_template = dedent(
            """\
            ### Stage 2: Option Selection

            Your task:
            - Select the single criterion option that best matches the text.
            - Optionally, choose an **alternative** option if you’re not completely certain.
            - If fully confident in your first choice, leave the alternative as an empty string.

            Available options:
            {criteria_option_names}

            ### Output format
            {format_instructions}
        """
        )

        second_stage_prompts = [
            second_stage_prompt_template.format(
                criteria_option_names=", ".join(
                    [option.name for option in criterion.options]
                ),
                format_instructions=second_stage_format_instructions,
            )
            for criterion in criteria
        ]

        second_stage_messages_list = [
            [
                *first_stage_message,
                self.get_ai_message_from_prompt(
                    first_stage_parsed_response, "assistant"
                ),
                self.get_ai_message_from_prompt(second_stage_prompt),
            ]
            for first_stage_message, first_stage_parsed_response, second_stage_prompt in zip(
                first_stage_messages,
                first_stage_unparsed_responses,
                second_stage_prompts,
            )
        ]

        second_stage_unparsed_responses = cast(
            list[str],
            self.inference_engine.infer(
                [
                    {
                        "source": second_stage_messages,
                        "data_classification_policy": ["public"],
                    }
                    for second_stage_messages in second_stage_messages_list
                ]
            ),
        )
        second_stage_output_parsed_responses, _ = parser.parse_and_repair(
            second_stage_unparsed_responses,
            [second_stage_model] * len(second_stage_unparsed_responses),
        )

        # Third stage (if required)
        second_stage_selected_option_list = [
            second_stage_output_parsed_response.model_dump()["selected_option"]
            for second_stage_output_parsed_response in second_stage_output_parsed_responses
        ]

        is_unsure_list = [
            second_stage_output_parsed_response.model_dump()[
                "alternative_selected_option"
            ]
            != ""
            for second_stage_output_parsed_response in second_stage_output_parsed_responses
        ]
        unsure_indexes: list[int] = []
        for i, (is_unsure, selected_option) in enumerate(
            zip(is_unsure_list, second_stage_selected_option_list)
        ):
            if is_unsure:
                unsure_indexes.append(i)

        third_stage_messages_list: list[list[dict[str, str]]] = []
        third_stage_format_instructions_list = []
        third_stage_models = []
        option_one_list = []
        option_two_list = []
        for unsure_index in unsure_indexes:
            option_one = second_stage_output_parsed_responses[
                unsure_index
            ].model_dump()["selected_option"]
            option_one_list.append(option_one)
            option_two = second_stage_output_parsed_responses[
                unsure_index
            ].model_dump()["alternative_selected_option"]
            option_two_list.append(option_two)

            third_stage_model = generate_dynamic_pydantic_model(
                model_name="third_stage_model",
                field_definitions=[
                    (
                        option_one,
                        str,
                        Field(
                            ...,
                            description=f"The argument for not evaluating the text as {option_one}",
                        ),
                        [],
                    ),
                    (
                        option_two,
                        str,
                        Field(
                            ...,
                            description=f"The argument for not evaluating the text as {option_two}",
                        ),
                        [],
                    ),
                ],
            )

            third_stage_models.append(third_stage_model)

            format_instructions = build_format_instructions(third_stage_model)
            third_stage_format_instructions_list.append(format_instructions)

            third_stage_prompt_template = dedent(
                """\
                ### Stage 3: Examining Alternatives

                You previously narrowed your choice to two options: **Option {option_one}** and **Option {option_two}**.

                Your task now is to critically assess **why the text should *not*** be assigned to each of these options.

                - Provide a brief rationale for why the text fails to match **Option {option_one}**.
                - Provide a brief rationale for why the text fails to match **Option {option_two}**.

                ### Output format
                {format_instructions}
            """,
            )

            third_stage_prompt = third_stage_prompt_template.format(
                option_one=option_one,
                option_two=option_two,
                format_instructions=format_instructions,
            )

            third_stage_messages = [
                *second_stage_messages_list[unsure_index],
                self.get_ai_message_from_prompt(
                    second_stage_unparsed_responses[unsure_index], "assistant"
                ),
                self.get_ai_message_from_prompt(third_stage_prompt),
            ]
            third_stage_messages_list.append(third_stage_messages)

        third_stage_unparsed_responses = cast(
            list[str],
            self.inference_engine.infer(
                [
                    {
                        "source": third_stage_messages,
                        "data_classification_policy": ["public"],
                    }
                    for third_stage_messages in third_stage_messages_list
                ]
            ),
        )

        # Fourth stage (if required)
        fourth_stage_messages_list = []
        fourth_stage_output_models = []
        for (
            option_one,
            option_two,
            third_stage_messages,
            third_stage_unparsed_response,
            unsure_index,
        ) in zip(
            option_one_list,
            option_two_list,
            third_stage_messages_list,
            third_stage_unparsed_responses,
            unsure_indexes,
        ):
            fourth_stage_output_model = generate_dynamic_pydantic_model(
                model_name="fourth_stage_model",
                field_definitions=[
                    (
                        "explanation",
                        str,
                        Field(
                            ...,
                            description="An explanation that clearly states why you selected that option",
                        ),
                        [],
                    ),
                    (
                        "selected_option",
                        str,
                        Field(
                            ...,
                            description=f"the selected option (either '{option_one}' or '{option_two}')",
                        ),
                        [],
                    ),
                ],
            )
            fourth_stage_output_models.append(fourth_stage_output_model)
            fourth_stage_format_instructions = build_format_instructions(
                fourth_stage_output_model
            )

            fourth_stage_prompt_template = dedent(
                """\
                ### Stage 4: Final Decision

                You have narrowed down your choices to two possible options:
                - **Option {option_one}**: {option_one_description}
                - **Option {option_two}**: {option_two_description}

                **Task**:
                1. Review the arguments *for* and *against* each option.
                2. Provide an explanation that describes, step by step, why you chose that option.
                3. Choose the one you believe fits best.

                **Context**:
                {context}

                **Text to evaluate**:
                {text_to_evaluate}

                ### Output format
                {format_instructions}
            """,
            )

            fourth_stage_prompt = fourth_stage_prompt_template.format(
                text_to_evaluate=predictions[unsure_index],
                context=str_context_variables_list[unsure_index],
                option_one_description=next(
                    iter(
                        o.description
                        for o in criteria[unsure_index].options
                        if o.name == option_one
                    )
                ),
                option_two_description=next(
                    iter(
                        o.description
                        for o in criteria[unsure_index].options
                        if o.name == option_two
                    )
                ),
                format_instructions=fourth_stage_format_instructions,
                option_one=option_one,
                option_two=option_two,
            )

            fourth_stage_messages: list[dict[str, str]] = [
                *third_stage_messages,
                self.get_ai_message_from_prompt(
                    third_stage_unparsed_response, "assistant"
                ),
                self.get_ai_message_from_prompt(fourth_stage_prompt),
            ]
            fourth_stage_messages_list.append(fourth_stage_messages)

        fourth_stage_unparsed_responses = cast(
            list[str],
            self.inference_engine.infer(
                [
                    {
                        "source": fourth_stage_messages,
                        "data_classification_policy": ["public"],
                    }
                    for fourth_stage_messages in fourth_stage_messages_list
                ]
            ),
        )

        fourth_stage_output_parsed_responses, _ = parser.parse_and_repair(
            unparsed_responses=fourth_stage_unparsed_responses,
            model_classes=fourth_stage_output_models,
        )

        fourth_stage_selected_option_list = [
            fourth_stage_output_parsed_response.model_dump()["selected_option"]
            for fourth_stage_output_parsed_response in fourth_stage_output_parsed_responses
        ]

        i = 0
        j = 0
        selected_options: list[str] = []
        explanations = []
        amount_of_stages = []
        full_messages_list = []

        for is_unsure in is_unsure_list:
            if is_unsure:
                selected_options.append(fourth_stage_selected_option_list[i])
                explanations.append(
                    fourth_stage_output_parsed_responses[i].model_dump()["explanation"]
                )
                full_messages_list.append(
                    fourth_stage_messages_list[i]
                    + [
                        self.get_ai_message_from_prompt(
                            fourth_stage_unparsed_responses[i], "assistant"
                        )
                    ]
                )
                amount_of_stages.append(4)
                i += 1
            else:
                selected_options.append(second_stage_selected_option_list[j])
                explanations.append(
                    first_stage_parsed_responses[j].model_dump()[
                        f"Arguments for {second_stage_selected_option_list[j]}"
                    ]
                )
                full_messages_list.append(
                    second_stage_messages_list[j]
                    + [
                        self.get_ai_message_from_prompt(
                            second_stage_unparsed_responses[j], "assistant"
                        )
                    ]
                )
                amount_of_stages.append(2)
            j += 1
        return [
            DirectInstanceResult(
                criteria=criterion,
                selected_option=selected_option,
                explanation=explanation,
                metadata={"Number of stages": str(stages_count)},
            )
            for selected_option, explanation, stages_count, criterion in zip(
                selected_options, explanations, amount_of_stages, criteria
            )
        ]
